#!/usr/bin/env python3.4
#
# mixd.py
#
# Copyright (c) 2014, DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


############
# Includes
############
import re
import os
import subprocess
from sys import argv
from configparser import SafeConfigParser, NoSectionError, NoOptionError
from logging import Formatter, getLogger, StreamHandler, DEBUG
from logging.config import fileConfig
import pyinotify
from pyinotify import WatchManager, Notifier, ProcessEvent
from multiprocessing import Process, Event
import signal


# default path to config file for production system
CONFFILE = '/services/etc/mixmod/mixd.conf'
MAIN_PROCESSES = '/services/inst/mixmod/bin/main-processes.py'


# noinspection PyUnboundLocalVariable
def process_config():
    global MAIN_PROCESSES
    """
    Processes the configuration file and checks for various options.
    """
    # Open and read config file
    config = SafeConfigParser({'create_directories': 'false',
                               'delete_input_file': 'false',
                               'mp-conf': '',
                               'file_regex': '.*'})

    try:
        config_file = argv[argv.index('--config') + 1]
    except ValueError:
        config_file = CONFFILE
    except IndexError:
        print('Config file name missing:')
        print(('Usage: %s [--config FILE]' % argv[0]))
        exit(1)
    if not config.read(config_file):
        print(('Unable to load config from file "{}". Does file "{}" exist?'.format(
            config_file, config_file)))
        exit(1)

    # configuring logging
    try:
        fileConfig(config_file)
        logger = getLogger('MixD (from config)')
    except NoSectionError:
        logger = getLogger('MixD (Default)')
        handler = StreamHandler()
        fmt_string = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        formatter = Formatter(fmt_string)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(DEBUG)

    try:
        MAIN_PROCESSES = config.get("main", "main_processes")
    except (NoSectionError, NoOptionError):
        pass

    watchers = list()

    for section in config.sections():
        if section[0:8].lower() == 'watcher:':
            watcher = dict()
            watcher["name"] = section[8:]

            # get folders to watch
            folders = list()
            for opt in config.options(section):
                if opt[0:12].lower() == "folder_plain":
                    folders.append({'path': config.get(section, opt), 'recursive': False})
                if opt[0:10].lower() == "folder_rec":
                    folders.append({'path': config.get(section, opt), 'recursive': True})
            watcher["folders"] = folders

            # config file (to be passed to main-processes)
            watcher["mp-conf"] = config.get(section, "mp-conf")

            try:
                watcher["pattern"] = re.compile(config.get(section, "file_regex"))
            except re.error:
                msg = "Invalid regular expression (file_regex) for {}".format(section)
                logger.critical(msg)
                exit(1)

            # regular expression to match directory names
            try:
                watcher["infile_regex"] = re.compile(config.get(section,
                                                                "infile_regex"))
            except NoOptionError:
                pass
            except re.error:
                msg = "Invalid regular expression (infile_regex) for %s" % section
                logger.critical(msg)
                exit(1)

            # gather output files for main-processes
            output = list()
            for opt in config.options(section):
                if opt[0:6].lower() == "output":
                    output.append(config.get(section, opt))
            watcher["output"] = output

            watcher['delete-after'] = config.getboolean(section, "delete_input_file")
            watcher['create_directories'] = config.getboolean(section, "create_directories")

            watchers.append(watcher)

    return logger, watchers


class Watcher(Process):
    """
    Process of the file importer.
    """

    def __init__(self, config, logger, do_stop):

        Process.__init__(self)
        self.logger = logger
        self.conf = config
        self.name = config['name']
        self.folders = config['folders']
        self.pattern = config['pattern']
        self.executer = None
        self.do_stop = do_stop

    def run(self):
        self.logger.debug('PID {}: Watcher'.format(self.pid))
        watchmanager = WatchManager()
        notifier = Notifier(watchmanager, self.HandleEvents(self.conf,
                            self.pattern, self.logger),
                            timeout=1000)
        mask = pyinotify.IN_CREATE | pyinotify.IN_MOVED_TO
        for folder in self.folders:
            watchmanager.add_watch(folder['path'], mask,
                                   rec=folder['recursive'],
                                   auto_add=True)
        while not self.do_stop.is_set():
            notifier.process_events()
            if notifier.check_events():
                notifier.read_events()
        self.stop()

    def stop(self):
        if self.executer is not None:
            self.logger.debug('Waiting for executer PID {} to finish...'.format(self.executer.pid))
            self.executer.join()
        self.logger.debug('Watcher PID {} exiting'.format(self.pid))

    class HandleEvents(ProcessEvent):
        def __init__(self, conf, pattern, logger):
            self.conf = conf
            self.pattern = pattern
            self.logger = logger

        def process_IN_MOVED_TO(self, event):
            self.process_IN_CREATE(event)

        def process_IN_CREATE(self, event):
            if self.pattern.match(event.name):
                matches = self.pattern.match(event.name)
                # compose arguments for main-processes
                self.executer = Executer(self.logger, self.conf,
                                    os.path.join(event.path, event.name),
                                    matches.groupdict())
                self.executer.run()


class Executer(Process):
    """
    Execute main-processes.py
    """

    def __init__(self, logger, conf, in_file, groupdict):
        super(Executer, self).__init__()
        self.logger = logger
        self.conf = conf
        self.in_file = in_file
        self.groupdict = groupdict

        infile_regex = self.conf.get("infile_regex", None)
        if infile_regex:
            matches = infile_regex.match(self.in_file)
            if matches:
                self.groupdict.update(matches.groupdict())

    def run(self):
        self.logger.debug('PID {}: Executer'.format(self.pid))
        try:
            arguments = [MAIN_PROCESSES, self.in_file]
            for outfile in self.conf['output']:
                try:
                    outfile_formatted = outfile.format(**self.groupdict)
                    if self.conf['create_directories']:
                        dirname = os.path.dirname(outfile_formatted)
                        if not os.path.exists(dirname):
                            os.makedirs(dirname)

                    arguments.append(outfile_formatted)
                except KeyError:
                    print(("{}".format(self.groupdict)))
                    msg = "Erroneous outputfile format '{}' for Watcher '{}'".format(outfile, self.conf['name'])
                    self.logger.critical(msg)
                    msg = "Watcher '{}' will be terminated!".format(self.conf['name'])
                    self.logger.critical(msg)
                    exit(1)

            if "mp-conf" in self.conf:
                arguments.append("--config")
                arguments.append(self.conf['mp-conf'])

            subprocess.check_call(arguments)
            # continue here if call was succesful
            # file has been processed succesfully
            if self.conf['delete-after']:
                try:
                    os.remove(self.in_file)
                except OSError:
                    pass

        except subprocess.CalledProcessError:
            msg = "Failed to process '%s'" % self.in_file  # handle errors
            self.logger.critical(msg)


def terminate(eventlist):
    logger = getLogger('MixD (from config)')
    for event in eventlist:
        event.set()
    logger.debug('terminating main')
    exit(1)

def main():
    # process configuration file
    (logger, watchers) = process_config()
    # create processes for each controller folder
    #
    processes = dict()
    do_stop_list = list()
    for watcher in watchers:
        do_stop = Event()
        processes[watcher['name']] = Watcher(watcher, logger, do_stop)
        do_stop_list.append(do_stop)
        processes[watcher['name']].start()
    signal.signal(signal.SIGTERM, lambda signum, frame: terminate(do_stop_list))

if __name__ == '__main__':
    main()
