#!/usr/bin/env python3.4
# nfgrep.py
#
# Copyright (c) 2014, 2015  DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from io import FileIO, BytesIO
from os import getcwd, listdir, path
from sys import argv, stdin
from os.path import isfile
from json import loads

from mixmod.modules.filter import Filter
from mixmod.errors import ConfigError, DataError
from mixmod.modules.netflow9importer import Netflow9Importer
from mixmod.modules.csdumpimporter import CSDumpImporter
from mixmod.modules.pcapimporter import PCapImporter
from mixmod.modules.nfcapimporter import NFCapImporter
from mixmod.modules.acdcjsonimporter import ACDCJSONImporter


TOOLTEXT = """
Usage: %s "POLICY" FILE|DIR [--verbose]
   POLICY: policy string in quotation marks (see Filter documentation)
   FILE|DIR: filename or directory name
-----------------------------------------------------------------------
This tool searches in files for a specific source IP address,
destination IP address, IP address, source port, destination port,
port, source AS, destination AS, AS, source country, destination
country, country, IPv4, IPv6, IP protocol number, TCP flags Syn / Fin /
Ack / Res, start time of the flow, end time of the flow, source net and
destination net. These can be specified by expressions that can be
combined by the logical operators AND and OR. Supported are the file
types Netflow V.9, CSDump, PCap, NFCap and the ACDC specific JSON.
Optionally the output of another command can be piped to this tool to be
searched through.
-----------------------------------------------------------------------
Examples:
%s "src ip 192.168.0.1" file.nf9
cat file.nf9 | %s "src ip 192.168.0.1"
"""


def main():
    if not stdin.isatty():
        stream = BytesIO()
        stream.write(stdin.read())
        if len(argv) == 2:
            policy, verbose = argv[1], False
        elif len(argv) == 3:
            policy, verbose = argv[1], True
        else:
            print(TOOLTEXT % (argv[0], argv[0], argv[0]))
            exit(1)
        search(policy, 'PIPED', stream, verbose)
        stream.close()
    else:
        if len(argv) == 3:
            policy, filename, verbose = argv[1], argv[2], False
        elif len(argv) == 4 and argv[3] == '--verbose':
            policy, filename, verbose = argv[1], argv[2], True
        else:
            print(TOOLTEXT % (argv[0], argv[0], argv[0]))
            exit(1)
        if path.isdir(path.join(getcwd(), filename)):
            traverse(policy, path.join(getcwd(), filename), verbose)
        else:
            file_open(policy, path.join(getcwd(), filename), verbose)


def traverse(policy, filepath, verbose):
    # read directory
    if verbose:
        print('DIR:', filepath)
    for d in listdir(filepath):
        file_in_dir = path.join(filepath, d)
        if path.isdir(file_in_dir):
            traverse(policy, file_in_dir, verbose)
        else:
            file_open(policy, file_in_dir, verbose)


def file_open(policy, filename, verbose):
    # read file
    if verbose:
        print('FILE:', filename)
    try:
        file_in = FileIO(filename, mode='r')
        stream = BytesIO()
        stream.write(file_in.readall())
        file_in.close()
    except IOError as e:
        if verbose:
            print(str(e))
        return None
    search(policy, filename, stream, verbose)


def search(policy, filename, stream, verbose):
    file_filter = Filter(policy=policy)
    importer_class, formatstring = getimporter(stream)
    stream.seek(0)
    if importer_class is not None:
        try:
            importer = importer_class(stream)
        except ConfigError as e:
            if verbose:
                print(str(e))
            return None
    else:
        if verbose:
            print(formatstring)
        return None

    for data in importer:
        try:
            file_filter.put(data)
        except DataError as e:
            if verbose:
                print('ERROR: %s' % str(e))
            continue
        for filtered in file_filter:
            print('%s:' % filename.rpartition('/')[2])
            if verbose:
                for key, item in filtered.items():
                    if key == 'type_of_service':
                        key = 'tos'
                    elif key == 'ip_protocol_number':
                        key = 'proto'
                    elif key == 'original_format':
                        key = 'format'
                    elif key == 'utc_first_packet':
                        key = 'utc_first'
                    elif key == 'utc_last_packet':
                        key = 'utc_last'
                    print('%s:%s' % (key, item))
                print('')
            else:
                print(filtered.get('ident'))

                if filtered.get('ip_version') == 4:
                    print("{:>8} {:>15}:{:<6} -> {:>15}:{:<5} AS{:<5}".format(
                        ('AS{}'.format(filtered.get('src_as'))),
                        filtered.get('src_ip_v4'),
                        filtered.get('src_port'),
                        filtered.get('dst_ip_v4'),
                        filtered.get('dst_port'),
                        filtered.get('dst_as')))

                elif filtered.get('ip_version') == 6:
                    print("{:>8} {:>39}:{:<6} -> {:>39}:{:<5} AS{:<5}".format(
                        ('AS{}'.format(filtered.get('src_as'))),
                        filtered.get('src_ip_v6'),
                        filtered.get('src_port'),
                        filtered.get('dst_ip_v6'),
                        filtered.get('dst_port'),
                        filtered.get('dst_as')))

                print('PROTO: %s START: %s END: %s' % (
                    filtered.get('ip_protocol_number'),
                    filtered.get('utc_first_packet'),
                    filtered.get('utc_last_packet')))


def getimporter(file_in):
    """
    This function expects a file name or a io.BytesIO object and
    returns the correct importer class and format string for this specific
    format. If the file does not exist, the file is not a regular file or the
    format is not supported by any of the importers, the tuple (None,
    errormessage) is returned. Otherwise the tuple (importer, formatstring) is
    returned.
    """
    if isinstance(file_in, str):
        if not isfile(file_in):
            return None, 'Not a regular file.'
        try:
            f = FileIO(file_in, mode='r')
            stream = BytesIO()
            stream.write(f.readall())
            f.close()
        except IOError:
            return None, 'Unable to open or read file.'
    elif isinstance(file_in, BytesIO):
        stream = file_in
    else:
        return None, 'Invalid argument type.'
    stream.seek(0)
    byte_0 = int(hex(ord(stream.read(1))), 16)
    byte_1 = int(hex(ord(stream.read(1))), 16)
    if byte_0 == 0x00 and byte_1 == 0x09:
        formatstring = 'nf9'
        importer = Netflow9Importer
    elif ((byte_0 == 0x1a and byte_1 == 0xb2) or
          (byte_0 == 0xd4 and byte_1 == 0xc3)):
        formatstring = 'pcap'
        importer = PCapImporter
    elif byte_0 == 0x0c and byte_1 == 0xa5:
        stream.seek(464)
        try:
            byte_464 = int(hex(ord(stream.read(1))), 16)
        except TypeError:
            byte_464 = None
        stream.seek(284)
        try:
            byte_284 = int(hex(ord(stream.read(1))), 16)
        except TypeError:
            return None, 'Probably empty NFCap file'
        if byte_464 == 0x1f:
            formatstring = 'csdump'
            importer = CSDumpImporter
        elif byte_284 == 0x02:
            formatstring = 'nfcap'
            importer = NFCapImporter
        else:
            return None, 'File format not supported.'
    else:
        stream.seek(0)
        try:
            loads(stream.getvalue())
        except ValueError:
            return None, 'File format not supported.'
        formatstring = 'acdcjson'
        importer = ACDCJSONImporter
    if isinstance(file_in, str):
        stream.close()
    return importer, formatstring


if __name__ == '__main__':
    main()
