#!/usr/bin/env python3.4
#
# update_geoip2_db.py
#
# Copyright (c) 2014, DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import tarfile
from urllib.request import urlopen
from urllib.error import URLError
from io import BytesIO
from os import rename
from configparser import SafeConfigParser, NoSectionError, NoOptionError
from logging import Formatter, getLogger, StreamHandler, WARNING, INFO
from sys import stderr, argv
from logging.config import fileConfig


# default database file
DEFAULT_GEOIP2_DB = '/services/inst/mixmod/data/GeoIP2.mmdb'

# url for geoip downloads
URL = 'http://geolite.maxmind.com/download/geoip/database/GeoLite2-City.tar.gz'

# default path to config file for production system
CONFFILE = '/services/etc/mixmod/mixmod.conf'


def download(logger, original_file):
    """
    Downloads and replaces the database files.
    """
    if not original_file:
        msg = 'Original filename expected, %s received.' % (original_file)
        logger.critical(msg)
        exit(1)
    renamed = True
    logger.info('Downloading file %s.' % URL)
    try:
        tar_ball = urlopen(URL)
    except URLError as e:
        msg = 'Unable to get file %s: %s' % (URL, str(e))
        logger.critical(msg)
    else:
        tar_file_obj = BytesIO(tar_ball.read())
        tar = tarfile.open(mode='r:gz', fileobj=tar_file_obj)
        # get name of the mmdb file
        mmdb_name = next(filter(lambda x: '.mmdb' in x, tar.getnames()))
        # only extract the mmdb file to a byte object
        mmdb_file_obj = tar.extractfile(mmdb_name)
        logger.info('Renaming file "%s" to "%s".' % (
            original_file, original_file + '_old'))
        try:
            rename(original_file, original_file + '_old')
        except OSError as e:
            renamed = False
        logger.info('Saving downloaded database file to "%s".' % original_file)
        try:
            with open(original_file, 'wb') as f:
                f.write(mmdb_file_obj.read())
        except IOError:
            msg = 'Unable to write file %s.' % original_file
            logger.critical(msg)
            if renamed:
                rename(original_file + '_old', original_file)
        finally:
            mmdb_file_obj.close()
            tar.close()
            tar_file_obj.close()
            tar_ball.close()


def get_config():
    """
    Processes the configuration file and checks for options.
    """
    config = SafeConfigParser()
    if '--help' in argv:
        print('Usage: %s [--verbose][--config FILE][--help]' % argv[0])
        exit(0)
    try:
        config_file = argv[argv.index('--config')+1]
    except ValueError:
        config_file = CONFFILE
        msg = '%s: No config file specified: Using %s.\n' % (
            argv[0], config_file)
        stderr.write(msg)
    except IndexError:
        stderr.write('Config file name missing:')
        stderr.write('Usage: %s [--verbose][--config FILE][--help]' % argv[0])
        exit(1)
    if not config.read(config_file):
        msg = '%s: Unable to load config from file. Does "%s" exist?\n' % (
            argv[0], config_file)
        stderr.write(msg)
        exit(1)
    # configuring logging
    try:
        fileConfig(config_file)
        logger = getLogger('MixModMain (from config)')
    except (NoOptionError, NoSectionError) as e:
        logger = getLogger('MixModMain (Default)')
        handler = StreamHandler()
        fmt_string = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        formatter = Formatter(fmt_string)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(WARNING)
        msg = 'Could not load config from file: %s. Using default logger.' % (
            str(e))
        logger.warning(msg)
    if '--verbose' in argv:
        logger.setLevel(INFO)
    db_file = None
    # get GeoIP database filenames
    for section in config.sections():
        if config.has_option(section, 'GeoIP2-DB-File'):
            db_file = config.get(section, 'GeoIP2-DB-File')
    if db_file is None:
        msg = 'Could not find option "GeoIP2-DB-File" in config file. '
        msg += 'Using default file %s instead.' % DEFAULT_GEOIP2_DB
        logger.warning(msg)
        db_file = DEFAULT_GEOIP2_DB

    return (logger, db_file)


if __name__ == '__main__':
    (logger, original_file) = get_config()
    download(logger, original_file)
