# acdcjsonexporter.py
#
# Copyright (c) 2015  DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import datetime
import json
import os

import jsonschema
import netaddr

from mixmod import rfc3339
from mixmod.errors import ConfigError, DataError
from mixmod.exporter import FileExporter

# possible report categories
ACDC_CATEGORIES = [
    "eu.acdc.attack",
    "eu.acdc.bot",
    "eu.acdc.botnet",
    "eu.acdc.c2_server",
    "eu.acdc.fast_flux",
    "eu.acdc.malicious_uri",
    "eu.acdc.malware",
    "eu.acdc.minimal",
    "eu.acdc.spam_campaign",
    "eu.acdc.vulnerable_uri"
]

DEFAULT_SCHEMA_DIR = "/services/etc/mixmod/schemata/"
DEFAULT_REPORT_TYPE = 'This dataset has been automatically generated by Mixmod'
DEFAULT_CONFIDENCE_LEVEL = 0.0
REQUIRE_SUBCAT = ['eu.acdc.attack', 'eu.acdc.bot', 'eu.acdc.botnet',
                  'eu.acdc.c2_server', 'eu.acdc.malicious_uri',
                  'eu.acdc.spam_campaign']


class ACDCJSONConverter(object):
    """Base exporter for mixmod data to ACDC JSON."""

    def __init__(self,
                 schema_dir=DEFAULT_SCHEMA_DIR,
                 default_report_category=None,
                 default_report_type=DEFAULT_REPORT_TYPE,
                 **kwargs):
        """Initialize ACDC base exporter.

        No positional arguments are consumed.

        The following keyword arguments are handled:
        @param schema_dir: The location of the ACDC schemata, default
            DEFAULT_SCHEMA_DIR
        @type schema_dir: string
        @param default_report_category: The default category to use for
            reports. If this is None, reports without a category are not
            exported. The default is None.
        @type default_report_category: string
        @param default_report_type: The default report type to use in absence
            of other values in the data, default DEFAULT_REPORT_TYPE
        @type default_report_type: string
        """
        self.schema_dir = schema_dir
        self.schema_version = 2
        self.load_schemata()
        self.default_report_category = default_report_category
        self.default_report_type = default_report_type
        super(ACDCJSONConverter, self).__init__(**kwargs)

    def load_schemata(self):
        """Load all schemata and check them."""
        self.schemata = dict()
        for category in ACDC_CATEGORIES:
            try:
                with open(os.path.join(self.schema_dir,
                                       'v{}'.format(self.schema_version),
                                       category + ".schema")) as f:
                    schema = json.load(f)
                    jsonschema.Draft4Validator.check_schema(schema)
                    self.schemata[category] = schema
            except IOError as e:
                msg = "failed to open schema for category '%s': %s" % (category, str(e))
                raise ConfigError(msg)
            except (ValueError, jsonschema.SchemaError) as e:
                msg = "invalid schema for category '%s': %s" % (category, str(e))
                raise ConfigError(msg)

    def convert_to_acdc(self, data):
        """Convert from internal representation dictionary types to what
        ACDC-JSON expects."""

        report_category = data.get('report_category',
                                   self.default_report_category)
        if report_category is None:
            return None
        if report_category not in ACDC_CATEGORIES:
            raise DataError('Invalid report_category: {}'
                            .format(report_category))
        acdc = {'report_category': report_category}

        for field in data:
            if field in TYPE_CONVERSIONS:
                try:
                    acdc[field] = TYPE_CONVERSIONS[field](data[field])
                except DataError as e:
                    raise DataError('Invalid field value for field {}: {} ({})'
                                    .format(field, data[field], e))
            else:
                acdc[field] = data[field]

        acdc['version'] = 2
        acdc['timestamp'] = acdc['utc_first_packet']
        if (
                'report_subcategory' not in acdc and
                report_category in REQUIRE_SUBCAT
        ):
            acdc['report_subcategory'] = 'other'
        if 'report_type' not in acdc:
            acdc['report_type'] = self.default_report_type
        if (
                'source_key' not in acdc or
                'source_value' not in acdc
        ):
            fill_source_key_and_value(acdc)
        if 'confidence_level' not in acdc:
            acdc['confidence_level'] = DEFAULT_CONFIDENCE_LEVEL

        if 'attachment' in acdc:
            acdc['additional_data'] = acdc['attachment']
            del acdc['attachment']

        properties = set(self.schemata[report_category]['properties'])
        for field in list(acdc.keys()):
            if field not in properties:
                del acdc[field]

        try:
            jsonschema.validate(acdc, self.schemata[report_category])
        except jsonschema.ValidationError as e:
            raise DataError('Schema error for report_category {}: {}'
                            .format(report_category, e))

        return acdc


class ACDCJSONExporter(FileExporter):
    """Export mixmod data to a file with ACDC JSON."""

    CONFIG = {
        "schema_dir": {
            "type": "string",
            "default": DEFAULT_SCHEMA_DIR
        },
        "default_report_category": {
            "type": "string",
            "default": None
        },
        "default_report_type": {
            "type": "string",
            "default": DEFAULT_REPORT_TYPE
        }
    }

    def __init__(self, *args, **kwargs):
        """Initialize ACDC base exporter.

        No positional arguments are consumed.

        The following keyword arguments are handled:
        @param schema_dir: The location of the ACDc schemata, default
            DEFAULT_SCHEMA_DIR
        @type schema_dir: string
        @param default_report_category: The default category to use for
            reports. If this is None, reports without a category are not
            exported. The default is None.
        @type default_report_category: string
        @param default_report_type: The default report type to use in absence
            of other values in the data, default DEFAULT_REPORT_TYPE
        @type default_report_type: string
        """
        schema_dir = kwargs.pop('schema_dir', DEFAULT_SCHEMA_DIR)
        default_report_category = kwargs.pop(
            'default_report_category', None)
        default_report_type = kwargs.pop(
            'default_report_type', DEFAULT_REPORT_TYPE)
        self.converter = ACDCJSONConverter(
            schema_dir=schema_dir,
            default_report_category=default_report_category,
            default_report_type=default_report_type)
        self.stream_started = False
        super(ACDCJSONExporter, self).__init__(*args, **kwargs)

    def put(self, data):
        """Export a mixmod data item.

        Only records which have the 'report_category' field with an
        entry in ACDC_CATEGORIES and which adhere to the matching
        schema will be added. Fields that are not part of the schema
        will be removed.

        """
        acdc = self.converter.convert_to_acdc(data)
        if acdc is None:
            return

        if not self.stream_started:
            self.fobj.write(b'[')
            self.stream_started = True
        else:
            self.fobj.write(b', ')

        self.fobj.write(str(acdc).encode('utf-8'))
        #json.dump(acdc, self.fobj)

    def close(self):
        """Finish emitting data.

        This does NOT close the file object."""
        if not self.stream_started:
            # Ensure a valid JSON list as the stream might not have started due
            # to missing report categories.
            self.fobj.write(b'[')
        self.fobj.write(b"]\n")


def from_datetime(dt):
    """Convert a datetime.datetime to a string according to RfC 3339
e.g. datetime.datetime(2015, 3, 14, 14, 30) -> '2015-03-14T14:30:00Z'

Cf. http://json-schema.org/latest/json-schema-validation.html#anchor108
and http://tools.ietf.org/html/rfc3339#section-5.6

@param dt: the datetime object to convert
@type dt: datetime.datetime
@return: date-time representation as per RfC 3339
@rtype: str
@raise DataError: the parameter was not a datetime.datetime instance
    """
    if dt is None:
        return None
    if isinstance(dt, datetime.datetime):
        return rfc3339.datetimetostr(dt)
    else:
        raise DataError('DateTime expected, but found {}'.format(type(dt)))


def from_ipaddr(ip):
    """Convert an IP v4 or v6 address object to a string
e.g. IPAddress('10.2.1.3') -> '10.2.1.3'

@param ip: IP address object to convert
@type ip: netaddr.IPAddress
@return: string representation in dotted-decimal or IPv6 notation
@rtype: str
@raise DataError: the parameter was not a netaddr.IPAddress instance
"""
    if ip is None:
        return None
    if isinstance(ip, netaddr.IPAddress):
        return str(ip)
    else:
        raise DataError('IPAddress expected, but found {}'.format(type(ip)))


def from_privacy(i):
    """Convert an integer referring to the used privacy method to a string
e.g. 2 -> 'pseudo'

@param i: integer reflecting used privacy method to convert
@type i: int
@return: string representation
@rtype: str
"""
    privacymap = {
        0: 'plain',
        1: 'pseudo',
        2: 'anon'
    }
    if i is None:
        return None
    try:
        return privacymap[int(i)]
    except:
        raise DataError('int referring to privacy method expected, but found {}'.format(i))



# The TYPE_CONVERSIONS dictionary contains a reference to the conversion functions defined above for
# each intermediate format field name that needs to be converted before being exported by the JSON
# export mechanism. For example, IP addresses are represented as netaddr.IPAddress instances in the
# Mixmod internal dictionary, but need to be exported as plain text according to the ACDC-JSON schema.
TYPE_CONVERSIONS = {
    'c2_ip_v4': from_ipaddr,
    'c2_ip_v6': from_ipaddr,
    'dst_ip_v4': from_ipaddr,
    'dst_ip_v6': from_ipaddr,
    'src_ip_v4': from_ipaddr,
    'src_ip_v6': from_ipaddr,
    'reported_at': from_datetime,
    'utc_first_packet': from_datetime,
    'src_mode': from_privacy,
    'dst_mode': from_privacy,
    'c2_mode': from_privacy
}


def fill_source_key_and_value(data):
    """Find and return the "source_key" and "source_value" entries from a given data dictionary

As these entries depend on the "report_category", we do this in a separate function. Note that when
this function is called, data has already undergone the type conversions above, i.e. all IP
addresses are already converted to strings. This function relies on that.

@param data: the Mixmod intermediate format dictionary to evaluate
@type data: dict[str, X]
@return: a tuple of "source_key", "source_value"
@rtype: (str, str)
@raise DataError: A sensible value for "source_key" or "source_value" could not be found
    """
    if data['report_category'] in ['eu.acdc.attack',
                                   'eu.acdc.bot',
                                   'eu.acdc.minimal']:
        data["source_key"] = "ip"
        try:
            data["source_value"] = data['src_ip_v4']
        except KeyError:
            try:
                data["source_value"] = data['src_ip_v6']
            except KeyError:
                raise DataError('No attack source IP in dataset')
    elif data['report_category'] == 'eu.acdc.botnet':
        data["source_key"] = 'botnet'
    elif data['report_category'] == 'eu.acdc.c2_server':
        data["source_key"] = 'ip'
        try:
            data["source_value"] = data['c2_ip_v4']
        except KeyError:
            try:
                data["source_value"] = data['c2_ip_v6']
            except KeyError:
                raise DataError('No c2 server IP in dataset')
    elif data['report_category'] == 'eu.acdc.fast_flux':
        data["source_key"] = 'uri'
    elif data['report_category'] == 'eu.acdc.malicious_uri':
        data["source_key"] = 'uri'
    elif data['report_category'] == 'eu.acdc.malware':
        data["source_key"] = 'malware'
    elif data['report_category'] == 'eu.acdc.spam_campaign':
        data["source_key"] = 'subject'
    elif data['report_category'] == 'eu.acdc.vulnerable_uri':
        data["source_key"] = 'uri'
    else:
        raise DataError('Could not determine source key and value for report_category: {}'
                        .format(data['report_category']))
