# acdcjsonimporter.py
#
# Copyright (c) 2015 DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from mixmod import importer
from mixmod.errors import ConfigError, DataError
from mixmod import rfc3339

import os
import netaddr
import json
import jsonschema
import unicodedata


# possible report categories

ACDC_CATEGORIES = [
    "eu.acdc.attack",
    "eu.acdc.bot",
    "eu.acdc.fast_flux",
    "eu.acdc.malware",
    "eu.acdc.spam_campaign",
    "eu.acdc.botnet",
    "eu.acdc.c2_server",
    "eu.acdc.malicious_uri",
    "eu.acdc.vulnerable_uri"
]

DEFAULT_SCHEMA_DIR = "/services/etc/mixmod/schemata/"


class ACDCJSONImporter(importer.FileImporter):
    CONFIG = {
        "schema_dir": {
            "type": "string",
            "default": DEFAULT_SCHEMA_DIR
        }
    }

    def __init__(self, fobj, schema_dir=DEFAULT_SCHEMA_DIR, **conf):
        super(ACDCJSONImporter, self).__init__(fobj, **conf)
        self.iterator = self._parse()
        self.schema_dir = schema_dir

        self.schemata = dict()
        versions = os.listdir(self.schema_dir)
        for category in ACDC_CATEGORIES:
            for version in versions:
                try:
                    with open(os.path.join(self.schema_dir, version,
                                           category + ".schema")) as f:
                        schema = json.load(f)
                        jsonschema.Draft4Validator.check_schema(schema)
                        self.schemata[(category, int(version[1:]))] = schema
                except IOError as e:
                    msg = "Failed to open schema for category '%s': %s." % (category, str(e))
                    raise ConfigError(msg)
                except (ValueError, jsonschema.SchemaError) as e:
                    msg = "Invalid schema for category '%s': %s" % (category, str(e))
                    raise ConfigError(msg)

    def get(self):
        try:
            return next(self.iterator)
        except StopIteration:
            return None

    def _parse(self):
        try:
            jsondata = json.loads(self.fobj.getvalue().decode(
                'utf-8').replace("'", '"'))
        except TypeError as e:
            msg = "Unable to parse input data: %s" % str(e)
            raise DataError(msg)

        # TODO - rename fields that must be renamed for ACDC JSON
        # - ACDC "additional_data" <-> Mixmod "attachment"/"attachment_type"

        # check if it is a single report or several reports
        if isinstance(jsondata, dict):
            # single report
            # validate input before yielding
            try:
                mm_dict = self._parse_single(jsondata)
                yield mm_dict
            except jsonschema.ValidationError as e:
                msg = "Data doesnt adhere to '%s' schema: %s" % (jsondata["report_category"], str(e))
                raise DataError(msg)
            except KeyError:
                msg = "'report_category' or 'version' field missing in report"
                raise DataError(msg)

        elif isinstance(jsondata, list):
            # list of reports
            for report in jsondata:
                try:
                    # validate data against schema
                    mm_dict = self._parse_single(report)
                    yield mm_dict
                except jsonschema.ValidationError as e:
                    msg = "Data doesnt adhere to '%s' schema: %s" % (report["report_category"], str(e))
                    raise DataError(msg)
                except KeyError:
                    msg = "At least one report does not contain the required 'report_category' field!"
                    raise DataError(msg)

    def _parse_single(self, raw_data):
        jsonschema.validate(
            raw_data, self.schemata[(raw_data["report_category"],
                                     raw_data["version"])])
        for field in raw_data.keys():
            # delete all fields that are not in the schema
            if field not in self.schemata[
                (raw_data["report_category"], raw_data["version"])]["properties"].keys():
                del raw_data[field]
            # convert types as defined in Mixmod intermediate format spec
            if field in type_conversions:
                raw_data[field] = type_conversions[field](raw_data[field])
            if isinstance(raw_data[field], str):
                raw_data[field] = unicodedata.normalize("NFKD", raw_data[field])
                # normalized = unicodedata.normalize("NFKD", raw_data[field])
                # raw_data[field] = normalized.encode("ascii", "ignore")
        if 'timestamp' in raw_data:
            raw_data['utc_first_packet'] = raw_data['timestamp']
        return raw_data


def to_datetime(dt):
    if dt is None:
        return None
    if isinstance(dt, str):
        return rfc3339.parse_datetime(dt)
    else:
        raise DataError('str expected, but found {}'.format(type(dt)))


def to_ipaddr(ip):
    if ip is None:
        return None
    if isinstance(ip, str):
        return netaddr.IPAddress(ip)
    else:
        raise DataError('str expected, but found {}'.format(type(ip)))

def to_privacy(s):
    privacymap = {
        'plain': 0,
        'anon': 1,
        'pseudo': 2
    }
    return privacymap.get(s, 0)

type_conversions = {
    'c2_ip_v4': to_ipaddr,
    'c2_ip_v6': to_ipaddr,
    'dst_ip_v4': to_ipaddr,
    'dst_ip_v6': to_ipaddr,
    'reported_at': to_datetime,
    'src_ip_v4': to_ipaddr,
    'src_ip_v6': to_ipaddr,
    'timestamp': to_datetime,
    'src_mode': to_privacy,
    'dst_mode': to_privacy
}
