# csdumpexporter.py
#
# Copyright (c) 2015 DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from datetime import datetime
import io

from mixmod import exporter
from mixmod.errors import DataError
from mixmod import csdumpformat


EPOCH = datetime(1970, 1, 1)
FWZBLOCK_FIELDS = ["device_id", "sector", "organisation_id"]
IDSBLOCK_FIELDS = ["signature_id", "organisation_id", "impact"]
MVBLOCK_FIELDS = ["uri", "sha1", "vuln_id", "md5", "mv_len",
                  "exploid_id", "event_count", "message_digest_type"]
ARGOSBLOCK_FIELDS = ["arg_version_major", "arg_version_minor",
                     "alarm_type", "argos_ethernet_type", "argos_vm_id",
                     "argos_id", "cpu_eip", "cpu_faulty_eip", "cpu_eax",
                     "cpu_ebx", "cpu_ecx", "cpu_edx", "cpu_ebp", "cpu_edi",
                     "cpu_esi", "cpu_esp", "cpu_eflags"]
REVBLOCK_FIELDS = ["rev_octets", "rev_tcpflags", "rev_tos"]
PSEUDOBLOCK_FIELDS = ["srcip_privacy_mode", "dstip_privacy_mode"]
ADBLOCK_FIELDS = ["attachment"]

FLAG_IPV6_ADDR = (1 << 0)
FLAG_PKG_64 = (1 << 1)
FLAG_BYTES_64 = (1 << 3)
FLAG_PSEUDOBLOCK = (1 << 23)
FLAG_REVBLOCK = (1 << 24)
FLAG_ARGOS = (1 << 25)
FLAG_MV = (1 << 27)
FLAG_FWZ = (1 << 28)
FLAG_NIDS_S = (1 << 29)
FLAG_ADD_DATA = (1 << 30)


class CSDumpExporter(exporter.Exporter):
    CONFIG = {
        "standard_direction": {"type": "int", "default": "0"},
        "standard_environment": {"type": "int", "default": "0"},
        "standard_location": {"type": "int", "default": "2"},
        "standard_data_class": {"type": "int", "default": "0"},
        "standard_data_format": {"type": "int", "default": "0"},
        "standard_organisation": {"type": "int", "default": "0"}
    }

    def __init__(self, fobj,
                 standard_direction=0,  # ingress
                 standard_environment=0,  # network
                 standard_location=2,  # internet
                 standard_data_class=0,  # flow
                 standard_data_format=0,  # unused
                 standard_organisation=0,  # none
                 **conf):
        super(CSDumpExporter, self).__init__(**conf)
        self.fobj = fobj
        self.standard_organisation = standard_organisation
        self.csfile = CSFile(fobj,
                             header_defaults={"flags": 0,
                                              "ident": "",
                                              "environment": standard_environment,
                                              "location": standard_location,
                                              "data_class": standard_data_class,
                                              "data_format": standard_data_format,
                                              "src_mode": 0,
                                              "dst_mode": 0},
                             standard_direction=int(standard_direction))

    def put(self, data):
        record = self.translate_record(data)
        self.csfile.add_record(record)

    def close(self):
        self.csfile.close()

    def translate_record(self, data):
        data = data.copy()
        data.setdefault("organisation", self.standard_organisation)
        return data


class CSFile(object):
    def __init__(self, fobj, header_defaults, standard_direction):
        self.fobj = fobj
        self.file_header = CSFileHeader(fobj, defaults=header_defaults)
        self.file_header.write()
        self.stat_record = CSStatsRecord(fobj)
        self.stat_record.write()
        self.current_block = None
        self.standard_direction = standard_direction

    def add_record(self, record):
        self.file_header.add_record(record)
        self.stat_record.add_record(record)
        if (
                self.current_block is not None and
                self.current_block.organisation != record["organisation"]
        ):
            self.current_block.close()
            self.current_block = None
        if self.current_block is None:
            self.current_block = CSBlockHeader(
                self.fobj,
                record["organisation"],
                standard_direction=self.standard_direction
            )
            self.current_block.write()
            self.file_header.add_block()
        self.current_block.add_record(record)

    def close(self):
        self.file_header.close()
        self.stat_record.close()
        if self.current_block is not None:
            self.current_block.close()


class CSFileSection(object):
    def __init__(self, fobj):
        self.fobj = fobj
        self.position = fobj.tell()

    def write(self):
        pass

    def close(self):
        old_position = self.fobj.tell()
        self.fobj.seek(self.position)
        self.write()
        self.fobj.seek(old_position)


class CSFileHeader(CSFileSection):
    def __init__(self, fobj, defaults):
        super(CSFileHeader, self).__init__(fobj)

        self.defaults = defaults

        self.magic = csdumpformat.MAGIC
        self.version = 1
        self.numblocks = 0

        self.flags = None
        self.ident = None
        self.environment = None
        self.location = None
        self.data_class = None
        self.data_format = None
        self.src_mode = None
        self.dst_mode = None

    def add_block(self):
        self.numblocks += 1

    def _set_field(self, record, field):
        if field not in record:
            return
        current = getattr(self, field)
        new = record[field]
        if current is None:
            setattr(self, field, record[field])
        elif current != new:
            setattr(self, field, self.defaults[0])

    def add_record(self, record):
        for field in ["flags", "ident", "environment",
                      "location", "data_class", "data_format",
                      "src_mode", "dst_mode"]:
            self._set_field(record, field)
        self._set_field(record, "flags")
        self._set_field(record, "ident")
        self._set_field(record, "environment")
        self._set_field(record, "location")
        self._set_field(record, "data_class")
        self._set_field(record, "data_format")
        self._set_field(record, "src_mode")
        self._set_field(record, "dst_mode")

    def write(self):
        if self.ident is not None:
            self.ident = self.ident.encode('utf-8')
        data = csdumpformat.FILEHEADER.pack(
            self._ifnone(self.magic, 0),
            self._ifnone(self.version, 0),
            self._ifnone(self.flags, 0),
            self._ifnone(self.numblocks, 0),
            self._ifnone(self.ident, b''),
            self._ifnone(self.environment, 0),
            self._ifnone(self.location, 0),
            self._ifnone(self.data_class, 0),
            self._ifnone(self.data_format, 0),
            self._ifnone(self.src_mode, 0),
            self._ifnone(self.dst_mode, 0)
        )
        self.fobj.write(data)

    def _ifnone(self, value, default):
        if value is None:
            return default
        else:
            return value


class CSStatsRecord(CSFileSection):
    def __init__(self, fobj):
        super(CSStatsRecord, self).__init__(fobj)
        self.stats = {"num_flows": 0,
                      "num_bytes": 0,
                      "num_packets": 0,
                      "num_flows_tcp": 0,
                      "num_flows_udp": 0,
                      "num_flows_icmp": 0,
                      "num_flows_other": 0,
                      "num_flows_tcp_impact_info": 0,
                      "num_flows_tcp_impact_low": 0,
                      "num_flows_tcp_impact_medium": 0,
                      "num_flows_tcp_impact_high": 0,
                      "num_flows_udp_impact_info": 0,
                      "num_flows_udp_impact_low": 0,
                      "num_flows_udp_impact_medium": 0,
                      "num_flows_udp_impact_high": 0,
                      "num_flows_icmp_impact_info": 0,
                      "num_flows_icmp_impact_low": 0,
                      "num_flows_icmp_impact_medium": 0,
                      "num_flows_icmp_impact_high": 0,
                      "num_flows_other_impact_info": 0,
                      "num_flows_other_impact_low": 0,
                      "num_flows_other_impact_medium": 0,
                      "num_flows_other_impact_high": 0,
                      "num_flows_tcp_rated": 0,
                      "num_flows_udp_rated": 0,
                      "num_flows_icmp_rated": 0,
                      "num_flows_other_rated": 0,
                      "num_bytes_tcp": 0,
                      "num_bytes_udp": 0,
                      "num_bytes_icmp": 0,
                      "num_bytes_other": 0,
                      "num_bytes_rated": 0,
                      "num_packets_tcp": 0,
                      "num_packets_udp": 0,
                      "num_packets_icmp": 0,
                      "num_packets_other": 0,
                      "num_packets_rated": 0}
        self.first_seen = 0
        self.last_seen = 0
        self.first_seen_msec = 0
        self.sequence_failure = 0

    def add_record(self, record):
        self._add_unit_stats(record)

        if "utc_first_packet" in record:
            first_seen = self._datetime_to_unixtime(record["utc_first_packet"])
            if first_seen < self.first_seen or self.first_seen == 0:
                self.first_seen = first_seen
        if "date_last" in record:
            last_seen = self._datetime_to_unixtime(record["utc_last_packet"])
            if last_seen > self.last_seen:
                self.last_seen = last_seen

    def _datetime_to_unixtime(self, timestamp, msec=None):
        if msec is not None:
            timestamp = timestamp.replace(microsecond=msec)
        return (timestamp - EPOCH).total_seconds()

    def _add_unit_stats(self, record):
        num_bytes = record["byte_count"]
        num_packets = record["packet_count"]
        proto = self._proto_number_to_name(record.get("ip_protocol_number"))

        for (unit, value) in [("flows", 1),
                              ("bytes", num_bytes),
                              ("packets", num_packets)]:
            # unit in flows, bytes, packets
            # proto in tcp, udp, icmp, other
            self.stats["num_{unit}".format(unit=unit)] += value
            self.stats["num_{unit}_{proto}".format(unit=unit, proto=proto)] += value

            if "impact" in record:
                self.stats["num_{unit}_{proto}_rated".format(unit=unit, proto=proto)] += value
                rating = self._impact_to_rating(record["impact"])
                # rating in info, low, medium, high
                if unit == "flows":
                    field = "num_flows_{proto}_impact_{rating}".format(proto=proto, rating=rating)
                    self.stats[field] += value

    def _impact_to_rating(self, impact):
        try:
            return {0: 'info',
                    1: 'low',
                    2: 'medium',
                    3: 'high'}[impact]
        except KeyError:
            raise DataError("Invalid impact in mixmod data: {}".format(impact))

    def _proto_number_to_name(self, proto):
        return {1: 'icmp',
                6: 'tcp',
                17: 'udp'}.get(proto, "other")

    def _add_field(self, field, value):
        setattr(self, field, getattr(self, field) + value)

    def write(self):
        data = csdumpformat.STATSRECORD.pack(
            self.stats["num_flows"],
            self.stats["num_bytes"],
            self.stats["num_packets"],
            self.stats["num_flows_tcp"],
            self.stats["num_flows_udp"],
            self.stats["num_flows_icmp"],
            self.stats["num_flows_other"],
            self.stats["num_flows_tcp_impact_info"],
            self.stats["num_flows_tcp_impact_low"],
            self.stats["num_flows_tcp_impact_medium"],
            self.stats["num_flows_tcp_impact_high"],
            self.stats["num_flows_udp_impact_info"],
            self.stats["num_flows_udp_impact_low"],
            self.stats["num_flows_udp_impact_medium"],
            self.stats["num_flows_udp_impact_high"],
            self.stats["num_flows_icmp_impact_info"],
            self.stats["num_flows_icmp_impact_low"],
            self.stats["num_flows_icmp_impact_medium"],
            self.stats["num_flows_icmp_impact_high"],
            self.stats["num_flows_other_impact_info"],
            self.stats["num_flows_other_impact_low"],
            self.stats["num_flows_other_impact_medium"],
            self.stats["num_flows_other_impact_high"],
            self.stats["num_flows_tcp_rated"],
            self.stats["num_flows_udp_rated"],
            self.stats["num_flows_icmp_rated"],
            self.stats["num_flows_other_rated"],
            self.stats["num_bytes_tcp"],
            self.stats["num_bytes_udp"],
            self.stats["num_bytes_icmp"],
            self.stats["num_bytes_other"],
            self.stats["num_bytes_rated"],
            self.stats["num_packets_tcp"],
            self.stats["num_packets_udp"],
            self.stats["num_packets_icmp"],
            self.stats["num_packets_other"],
            self.stats["num_packets_rated"],
            int(self.first_seen),
            int(self.last_seen),
            int((self.first_seen - int(self.first_seen)) * 1000),
            int((self.last_seen - int(self.last_seen)) * 1000),
            self.sequence_failure
        )
        self.fobj.write(data)


class CSBlockHeader(CSFileSection):
    def __init__(self, fobj, organisation, standard_direction):
        super(CSBlockHeader, self).__init__(fobj)
        self.num_records = 0
        self.size = 0
        self.organisation = organisation
        self.id = csdumpformat.BLOCK_ID
        self.pad = 0
        self.standard_direction = standard_direction

    def add_record(self, record):
        self.num_records += 1
        rec = CSRecord(self.fobj, record, self.standard_direction)
        rec.write()
        self.size += rec.recordsize

    def write(self):
        data = csdumpformat.BLOCKHEADER.pack(
            self.num_records,
            self.size,
            self.organisation,
            self.id,
            self.pad
        )
        self.fobj.write(data)


class CSRecord(object):
    def __init__(self, fobj, record, standard_direction):
        self.fobj = fobj
        self.record = record
        self.payload = io.BytesIO()
        self.flags = 0
        self.mark = record.get("mark", 0)

        if "utc_first_packet" in record:
            first_secs = (record["utc_first_packet"] - EPOCH).total_seconds()
            self.first = int(first_secs)
            self.msec_first = int((first_secs - int(first_secs)) * 1000)
        if "utc_last_packet" in record:
            last_secs = (record["utc_last_packet"] - EPOCH).total_seconds()
            self.last = int(last_secs)
            self.msec_last = int((last_secs - int(last_secs)) * 1000)

        self.dir = record.get("direction", standard_direction)
        self.tcp = record.get("tcp_flags", 0)
        self.prot = record.get("ip_protocol_number", 0)
        self.tos = record.get("type_of_service", 0)
        self.input_ = record.get("snmp_index_in", 0)
        self.output = record.get("snmp_index_out", 0)
        self.srcport = record.get("src_port", 0)
        self.dstport = record.get("dst_port", 0)
        self.srcas = record.get("src_as", 0)
        self.dstas = record.get("dst_as", 0)
        self.meta = record.get("meta", 0)
        self.fill0 = 0
        self.meta_category = record.get("meta_category", 0)

        self.add_ip_address_block()
        self.add_fwz_block()
        self.add_ids_block()
        self.add_mv_block()
        self.add_argos_block()
        self.add_rev_block()
        self.add_pseudo_block()
        self.add_ad_block()

    @property
    def recordsize(self):
        return csdumpformat.COMMONRECORD.size + len(self.payload.getvalue())

    def write(self):
        data = csdumpformat.COMMONRECORD.pack(
                self.flags, self.recordsize, self.mark, self.msec_first,
                self.msec_last, self.first, self.last, self.dir, self.tcp,
                self.prot, self.tos, self.input_, self.output, self.srcport,
                self.dstport, self.srcas, self.dstas, self.meta, self.fill0,
                self.meta_category)
        self.fobj.write(data)
        self.fobj.write(self.payload.getvalue())

    def add_ip_address_block(self):
        num_packets = self.record["packet_count"]
        num_bytes = self.record["byte_count"]

        if self.record["ip_version"] == 4:
            src_ip = self.record["src_ip_v4"].value
            dst_ip = self.record["dst_ip_v4"].value
            data = csdumpformat.RECORD_V4BLOCK.pack(
                        src_ip,
                        dst_ip,
                        num_packets,
                        num_bytes)
            self.payload.write(data)
        else:
            self.flags |= FLAG_IPV6_ADDR | FLAG_PKG_64 | FLAG_BYTES_64
            src_ip = self.record["src_ip_v6"].value
            dst_ip = self.record["dst_ip_v6"].value
            data = csdumpformat.RECORD_V6BLOCK.pack(src_ip // (2**64),
                                                    src_ip % (2**64),
                                                    dst_ip // (2**64),
                                                    dst_ip % (2**64))
            self.payload.write(data)
            data = csdumpformat.RECORD_V6_PACKETS_AND_BYTES(num_packets,
                                                            num_bytes)
            self.payload.write(data)

    def add_fwz_block(self):
        if not all(field in self.record for field in FWZBLOCK_FIELDS):
            return
        self.flags |= FLAG_FWZ
        data = csdumpformat.FWZBLOCK.pack(
                    self.record["device_id"],
                    self.record["sector"],
                    self.record["add_data_type"],
                    0,
                    self.record["organisation_id"]
                )
        self.payload.write(data)

    def add_ids_block(self):
        if not all(field in self.record for field in IDSBLOCK_FIELDS):
            return
        self.flags |= FLAG_NIDS_S
        data = csdumpformat.IDSBLOCK.pack(
                    self.record["signature_id"],
                    self.record["origin"],
                    self.record["impact"],
                    0
                )
        self.payload.write(data)

    def add_mv_block(self):
        if not all(field in self.record for field in MVBLOCK_FIELDS):
            return
        self.flags |= FLAG_MV
        data = csdumpformat.MVBLOCK.pack(
                    self.record["uri"],
                    self.record["sha1"],
                    self.record["vuln_id"],
                    self.record["md5"],
                    self.record["mv_len"],
                    self.record["event_count"],
                    self.record["exploit_id"],
                    self.record["message_digest_type"],
                    0
                )
        self.payload.write(data)

    def add_argos_block(self):
        if not all(field in self.record for field in ARGOSBLOCK_FIELDS):
            return
        self.flags |= FLAG_ARGOS
        data = csdumpformat.ARGOSBLOCK.pack(
                    *[self.record[field] for field in ARGOSBLOCK_FIELDS]
                )
        self.payload.write(data)

    def add_rev_block(self):
        if not all(field in self.record for field in REVBLOCK_FIELDS):
            return
        self.flags |= FLAG_REVBLOCK
        data = csdumpformat.REVBLOCK.pack(
                    self.record["rev_packets"],
                    self.record["rev_octets"],
                    self.record["rev_tcpflags"],
                    self.record["rev_tos"]
                )
        self.payload.write(data)

    def add_pseudo_block(self):
        if not all(field in self.record for field in PSEUDOBLOCK_FIELDS):
            return
        self.flags |= FLAG_PSEUDOBLOCK
        data = csdumpformat.PSEUDOBLOCK.pack(
                    self.record["srcip_privacy_mode"],
                    self.record["dstip_privacy_mode"]
                )
        self.payload.write(data)

    def add_ad_block(self):
        if not all(field in self.record for field in ADBLOCK_FIELDS):
            return
        self.flags |= FLAG_ADD_DATA
        data = csdumpformat.ADBLOCK.pack(
                    self.record["attachment"]
                )
        self.payload.write(data)
