# csdumpimporter.py
#
# Copyright (c) 2015 DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from mixmod import importer
from datetime import timedelta, datetime
from mixmod.errors import FormatError
from mixmod import csdumpformat
import netaddr


# Magic to recognize endian type
CSDUMP_MAGIC = 0xA50C

# Constants
FLAG_IPV6_ADDR = (1 << 0)
FLAG_PKG_64 = (1 << 1)
FLAG_BYTES_64 = (1 << 2)
FLAG_PSEUDOBLOCK = (1 << 23)
FLAG_REVBLOCK = (1 << 24)
FLAG_ARGOS = (1 << 25)
# FLAG_BI_FlOW = (1 << 26)
FLAG_MV = (1 << 27)
FLAG_FWZ = (1 << 28)
FLAG_NIDS_S = (1 << 29)
FLAG_ADD_DATA = (1 << 30)


class CSDumpImporter(importer.FileImporter):
    """Import CarmentiS CSDump files."""

    CONFIG = {}

    def __init__(self, fobj, **conf):
        super(CSDumpImporter, self).__init__(fobj, **conf)
        self.mixmoddata = dict()
        self.rawdata = fobj.read()
        self.iterator = self._parse()

    def get(self):
        try:
            return next(self.iterator)
        except StopIteration:
            return None

    def _parse(self):
        """
        Parses the data in self.rawdata by using helper functions and generates
        mixmoddata dictionaries.
        """
        self._read_fileheader()
        self._read_statsrecord()
        self.offset = csdumpformat.FILEHEADER.size + \
            csdumpformat.STATSRECORD.size

        # Differences between flag (structure of this record) and size
        self.realSize = 0

        for ign in range(self.numblocks):
            self._parse_datablockheader()
            if self.idBlock == 31:
                for ign in range(self.numrecordsBlock):
                    self._parse_commonrecord()
                    yield self.mixmoddata
                    self.mixmoddata = dict()
            else:
                msg = "Blockheader: ID (%s) is not supported"
                raise FormatError(msg % (self.idBlock))

    def _read_fileheader(self):
        """
        Parses the main file header.
        """
        (magic, version, flags, numblocks, ident, environment, location,
         data_class, data_format, src_mode, dst_mode) = \
            csdumpformat.FILEHEADER.unpack_from(self.rawdata, 0)
        if magic != CSDUMP_MAGIC:
            msg = "Fileheader: Bad MAGIC in file"
            raise FormatError(msg)
        if version != 1:
            msg = "Fileheader: Bad Version in file. Only version 1 supported."
            raise FormatError(msg)

        self.numblocks = numblocks
        self.headerdata = {
            "magic": magic,
            "version": version,
            "ident": ident.decode('utf-8').replace("\x00", ""),
            "flags": flags,
            "environment": environment,
            "location": location,
            "data_class": data_class,
            "data_format": data_format,
            "src_mode": src_mode,
            "dst_mode": dst_mode}

    def _read_statsrecord(self):
        """
        Parses statistic records.
        """
        self.numflows = {}
        self.numbytes = {}
        self.numpackets = {}
        (self.numflows["total"],
         self.numbytes["total"],
         self.numpackets["total"],
         self.numflows["tcp"],
         self.numflows["udp"],
         self.numflows["icmp"],
         self.numflows["other"],
         self.numflows["tcp_impact_info"],
         self.numflows["tcp_impact_low"],
         self.numflows["tcp_impact_medium"],
         self.numflows["tcp_impact_high"],
         self.numflows["udp_impact_info"],
         self.numflows["udp_impact_low"],
         self.numflows["udp_impact_medium"],
         self.numflows["udp_impact_high"],
         self.numflows["icmp_impact_info"],
         self.numflows["icmp_impact_low"],
         self.numflows["icmp_impact_medium"],
         self.numflows["icmp_impact_high"],
         self.numflows["other_impact_info"],
         self.numflows["other_impact_low"],
         self.numflows["other_impact_medium"],
         self.numflows["other_impact_high"],
         self.numflows["tcp_rated"],
         self.numflows["udp_rated"],
         self.numflows["icmp_rated"],
         self.numflows["other_rated"],
         self.numbytes["tcp"],
         self.numbytes["udp"],
         self.numbytes["icmp"],
         self.numbytes["other"],
         self.numbytes["rated"],
         self.numpackets["tcp"],
         self.numpackets["udp"],
         self.numpackets["icmp"],
         self.numpackets["other"],
         self.numpackets["rated"],
         self.first_seen,
         self.last_seen,
         self.msec_first,
         self.msec_last,
         self.sequence_failure) = csdumpformat.STATSRECORD.unpack_from(
            self.rawdata, csdumpformat.FILEHEADER.size)

    def _parse_datablockheader(self):
        """
        Parses the Datablockheader.
        """
        (numrecords, size, organisation, id, pad) = \
            csdumpformat.BLOCKHEADER.unpack_from(
                self.rawdata, self.offset)
        self.numrecordsBlock = numrecords
        self.sizeBlock = size
        self.organisation = organisation
        self.idBlock = id
        self.offset += csdumpformat.BLOCKHEADER.size

    def _parse_commonrecord(self):
        """
        Parses the records.
        """
        self.mixmoddata.update(self.headerdata)

        (flags, size, mark, msec_first, msec_last, first, last, dir,
         tcp_flags, prot, src_tos, input, output, srcport, dstport,
         srcas, dstas, meta, fill0, meta_category) = \
            csdumpformat.COMMONRECORD.unpack_from(
                self.rawdata, self.offset)

        self.mixmoddata["original_format"] = "csd"
        self.mixmoddata["organisation"] = self.organisation
        self.mixmoddata["mark"] = mark
        self.mixmoddata["utc_first_packet"] = datetime.utcfromtimestamp(first) + \
            timedelta(milliseconds=msec_first)
        self.mixmoddata["utc_last_packet"] = datetime.utcfromtimestamp(last) + \
            timedelta(milliseconds=msec_last)
        self.mixmoddata["direction"] = dir
        self.mixmoddata["tcp_flags"] = tcp_flags
        self.mixmoddata["ip_protocol_number"] = prot
        self.mixmoddata["type_of_service"] = src_tos
        self.mixmoddata["snmp_index_in"] = input
        self.mixmoddata["snmp_index_out"] = output
        self.mixmoddata["src_port"] = srcport
        self.mixmoddata["dst_port"] = dstport
        self.mixmoddata["src_as"] = srcas
        self.mixmoddata["dst_as"] = dstas
        self.mixmoddata["meta"] = meta
        self.mixmoddata["meta_category"] = meta_category
        self.offset += csdumpformat.COMMONRECORD.size
        self.realSize += csdumpformat.COMMONRECORD.size

        # if src_as = 0 then delete entry
        if 0 == self.mixmoddata["src_as"]:
            del self.mixmoddata["src_as"]

        # if dst_as = 0 then delete entry
        if 0 == self.mixmoddata["dst_as"]:
            del self.mixmoddata["dst_as"]

        # IPv6 block
        if flags & FLAG_IPV6_ADDR:
            self.mixmoddata["ip_version"] = 6
            (ip1a, ip1b, ip2a, ip2b) = \
                csdumpformat.RECORD_V6BLOCK.unpack_from(
                    self.rawdata, self.offset)

            (src_ip_v6, dst_ip_v6) = (ip1a * 18446744073709551616 + ip1b,
                                      ip2a * 18446744073709551616 + ip2b)
            src_ip_v6 = netaddr.IPAddress(src_ip_v6, version=6)
            dst_ip_v6 = netaddr.IPAddress(dst_ip_v6, version=6)
            self.mixmoddata["src_ip_v6"] = src_ip_v6
            self.mixmoddata["dst_ip_v6"] = dst_ip_v6
            self.offset += csdumpformat.RECORD_V6BLOCK.size
            self.realSize += csdumpformat.RECORD_V6BLOCK.size

        # IPv4 block
        else:
            self.mixmoddata["ip_version"] = 4
            (srcip, dstip, numpackets, numbytes) = \
                csdumpformat.RECORD_V4BLOCK.unpack_from(
                    self.rawdata, self.offset)

            src_ip_v4 = netaddr.IPAddress(srcip, version=4)
            dst_ip_v4 = netaddr.IPAddress(dstip, version=4)
            self.mixmoddata["src_ip_v4"] = src_ip_v4
            self.mixmoddata["dst_ip_v4"] = dst_ip_v4
            self.mixmoddata["packet_count"] = numpackets
            self.mixmoddata["byte_count"] = numbytes
            self.offset += csdumpformat.RECORD_V4BLOCK.size
            self.realSize += csdumpformat.RECORD_V4BLOCK.size

        # Counter block
        if flags & FLAG_PKG_64 and flags & FLAG_BYTES_64:
            (numpackets, numbytes) = \
                csdumpformat.RECORD_V6_PACKETS_AND_BYTES.unpack_from(
                    self.rawdata, self.offset)

            self.mixmoddata["packet_count"] = numpackets
            self.mixmoddata["byte_count"] = numbytes
            self.offset += csdumpformat.RECORD_V6_PACKETS_AND_BYTES.size
            self.realSize += csdumpformat.RECORD_V6_PACKETS_AND_BYTES.size

        # FWZ block
        if flags & FLAG_FWZ:
            (analyser, sector, add_data_type, fill, organisation_id) = \
                csdumpformat.FWZBLOCK.unpack_from(
                    self.rawdata, self.offset)

            self.mixmoddata["device_id"] = analyser
            self.mixmoddata["sector"] = sector
            self.mixmoddata["add_data_type"] = add_data_type
            self.mixmoddata["organisation_id"] = organisation_id
            self.offset += csdumpformat.FWZBLOCK.size
            self.realSize += csdumpformat.FWZBLOCK.size

        # IDS block
        if flags & FLAG_NIDS_S:
            (sid, origin, impact, fill1) = \
                csdumpformat.IDSBLOCK.unpack_from(
                    self.rawdata, self.offset)

            self.mixmoddata["signature_id"] = sid
            self.mixmoddata["organisation_id"] = origin
            self.mixmoddata["impact"] = impact
            self.offset += csdumpformat.IDSBLOCK.size
            self.realSize += csdumpformat.IDSBLOCK.size

        # AD block
        if flags & FLAG_ADD_DATA:
            (add_data) = \
                csdumpformat.ADBLOCK.unpack_from(
                    self.rawdata, self.offset)

            self.mixmoddata["attachment"] = add_data
            self.offset += csdumpformat.ADBLOCK.size
            self.realSize += csdumpformat.ADBLOCK.size

        # MV block
        if flags & FLAG_MV:
            (uri, sha1, vuln_id, md5, mv_len, event_count,
             exploit_id, md_type, fill8) = \
                csdumpformat.MVBLOCK.unpack_from(
                    self.rawdata, self.offset)

            self.mixmoddata["uri"] = uri
            self.mixmoddata["sha1"] = sha1
            self.mixmoddata["vuln_id"] = vuln_id
            self.mixmoddata["md5"] = md5
            self.mixmoddata["mv_len"] = mv_len
            self.mixmoddata["event_count"] = event_count
            self.mixmoddata["exploid_id"] = exploit_id
            self.mixmoddata["message_digest_type"] = md_type
            self.offset += csdumpformat.MVBLOCK.size
            self.realSize += csdumpformat.MVBLOCK.size

        # Argos block
        if flags & FLAG_ARGOS:
            (arg_version_major, arg_version_minor, arg_alert_type,
             argos_ethernet_type, argos_vm_id, argos_id, arg_cpu_eip,
             arg_cpu_faulty_eip, arg_cpu_eax, arg_cpu_ebx, arg_cpu_ecx,
             arg_cpu_edx, arg_cpu_ebp, arg_cpu_edi, arg_cpu_esi,
             arg_cpu_esp, arg_cpu_eflags) = \
                csdumpformat.ARGOSBLOCK.unpack_from(
                    self.rawdata, self.offset)

            self.mixmoddata["arg_version_major"] = arg_version_major
            self.mixmoddata["arg_version_minor"] = arg_version_minor
            self.mixmoddata["alarm_type"] = arg_alert_type
            self.mixmoddata["argos_ethernet_type"] = argos_ethernet_type
            self.mixmoddata["argos_vm_id"] = argos_vm_id
            self.mixmoddata["argos_id"] = argos_id
            self.mixmoddata["cpu_eip"] = arg_cpu_eip
            self.mixmoddata["cpu_faulty_eip"] = arg_cpu_faulty_eip
            self.mixmoddata["cpu_eax"] = arg_cpu_eax
            self.mixmoddata["cpu_ebx"] = arg_cpu_ebx
            self.mixmoddata["cpu_ecx"] = arg_cpu_ecx
            self.mixmoddata["cpu_edx"] = arg_cpu_edx
            self.mixmoddata["cpu_ebp"] = arg_cpu_ebp
            self.mixmoddata["cpu_edi"] = arg_cpu_edi
            self.mixmoddata["cpu_esi"] = arg_cpu_esi
            self.mixmoddata["cpu_esp"] = arg_cpu_esp
            self.mixmoddata["cpu_eflags"] = arg_cpu_eflags
            self.offset += csdumpformat.ARGOSBLOCK.size
            self.realSize += csdumpformat.ARGOSBLOCK.size

        # Rev block
        if flags & FLAG_REVBLOCK:
            (rev_packets, rev_octets, rev_tcpflags, rev_tos) = \
                csdumpformat.REVBLOCK.unpack_from(
                    self.rawdata, self.offset)

            self.mixmoddata["rev_packets"] = rev_packets
            self.mixmoddata["rev_octets"] = rev_octets
            self.mixmoddata["rev_tcpflags"] = rev_tcpflags
            self.mixmoddata["rev_tos"] = rev_tos
            self.offset += csdumpformat.REVBLOCK.size
            self.realSize += csdumpformat.REVBLOCK.size

        # Pseudo block
        if flags & FLAG_PSEUDOBLOCK:
            (srcip_privacy_mode, dstip_privacy_mode) = \
                csdumpformat.PSEUDOBLOCK.unpack_from(
                    self.rawdata, self.offset)

            self.mixmoddata["srcip_privacy_mode"] = srcip_privacy_mode
            self.mixmoddata["dstip_privacy_mode"] = dstip_privacy_mode
            self.offset += csdumpformat.PSEUDOBLOCK.size
            self.realSize += csdumpformat.PSEUDOBLOCK.size

        # Calculates the difference between self.offset and self.realSize
        if (size - self.realSize) != 0:
            self.offset += size - self.realSize

        self.realSize = 0
