# filter.py
#
# Copyright (c) 2015 DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from collections import deque
from datetime import datetime

import geoip2.database

from mixmod import dataprocessor
from mixmod.errors import ConfigError, DataError
from mixmod.netflowfilter import ip2int
from mixmod.nfparse import ParsingError, nfparse

# constant symbols of comparison
COMP = ['=', '==', '>', '<', 'eq', 'lt', 'gt']

# mapping from nfparse terms to mixmod terms
MAP = {
    'srcip': ('src_ip_v4', 'src_ip_v6'),
    'dstip': ('dst_ip_v4', 'dst_ip_v6'),
    'srcport': 'src_port',
    'dstport': 'dst_port',
    'srcas': 'src_as',
    'dstas': 'dst_as',
    'srccountry': 'src_country',
    'dstcountry': 'dst_country',
    'ipversion': 'ip_version',
    'prot': 'ip_protocol_number',
    'tcp_flags': 'tcp_flags',
    'start': 'utc_first_packet',
    'end': 'utc_last_packet'
}

# time zero of unix seconds
TIME_ZERO = datetime(1970, 1, 1, 00, 00, 00, 000000)


class Filter(dataprocessor.DataProcessor):
    """Filter mixmod data.

    Data can be filtered for IP, port, AS, country, ip version,
    protocol, TCP flags, time and network.

    """

    CONFIG = {
        "policy": {
            "type": "string",
            "default": "port 443"
        },
        "geoip2_db_file": {
            "type": "string",
            "default": None
        },
    }

    def __init__(self, policy, geoip2_db_file=None, **conf):
        """The policy argument should be a filter description like one of the
        following:

        src ip SRC_IP_ADDRESS
        dst ip DST_IP_ADDRESS
        ip IP_ADDRESS
        src as [COMP] SRC_AS_NUMBER
        dst as [COMP] DST_AS_NUMBER
        as [COMP] AS_NUMBER
        src port [COMP] SRC_PORT
        dst port [COMP] DST_PORT
        port [COMP] PORT
        src country XX (COUNTRY_CODE)
        dst country XX (COUNTRY_CODE)
        country XX (COUNTRY_CODE)
        ipv4
        ipv6
        proto IP_PROTOCOL_NUMBER
        flags {f,s,r,a}
        start [COMP] YYYY MM DD HH MM SS
        end [COMP] YYYY MM DD HH MM SS
        src net CIDR
        dst net CIDR
        AND
        OR
        ()
        with COMP = {= | == | > | < | eq | lt | gt}
        Example:
        ( src port > 1024 ) and ( dst ip 192.168.0.100 )
        src net 10.0.0.0/8 or start > 2013 12 30 23 51 59
        src country CN and flags sa

        """
        self.policy_string = self._replace_policy_time(policy)
        self.geoip2_db = geoip2_db_file

        try:
            self.policy = nfparse(self.policy_string)
        except (ParsingError, RuntimeError) as e:
            msg = 'Error while parsing policy string: %s.' % str(e)
            raise ConfigError(msg)
        self.test = self.policy.make_test()

        self.results = deque()

    def put(self, data):
        if 'country' in self.policy_string:
            self._get_country_codes(data)
        wrapped_data = Wrapper(data)
        try:
            result = self.test(wrapped_data)
        except DataError:
            pass
        else:
            if result:
                self.results.append(data)

    def get(self):
        try:
            return self.results.popleft()
        except IndexError:
            return None

    def _get_country_codes(self, data):
        """
        Gets and saves the country code of the src und dst ip address in the
        mixmod dataset.
        """
        try:
            data['ip_version']
        except KeyError:
            msg = 'Field "ip_version" does not exist in mixmod record.'
            raise DataError(msg)

        data['src_country'] = None
        data['dst_country'] = None

        if self.geoip2_db is None:
            msg = 'geoip2_db_file is not configured, cannot filter for country'
            raise ConfigError(msg)
        try:
            geo_ip = geoip2.database.Reader(self.geoip2_db)
        except:
            msg = 'Unable to load geoIP database %s.' % self.geoip2_db
            raise EnvironmentError(msg)
        if data['ip_version'] == 4:
            if 'src_ip_v4' in data:
                country = geo_ip.city(str(data['src_ip_v4'])).country.iso_code
                if country != '':
                    data['src_country'] = country
            if 'dst_ip_v4' in data:
                country = geo_ip.city(str(data['dst_ip_v4'])).country.iso_code
                if country != '':
                    data['dst_country'] = country
        elif data['ip_version'] == 6:
            if 'src_ip_v6' in data:
                ipv6 = str(data['src_ip_v6'])
                if ipv6 != '::':
                    country = geo_ip.city(ipv6).country.iso_code
                    if country != '':
                        data['src_country'] = country
            if 'dst_ip_v6' in data:
                ipv6 = str(data['dst_ip_v6'])
                if ipv6 != '::':
                    country = geo_ip.city(ipv6).country.iso_code
                    if country != '':
                        data['dst_country'] = country
        else:
            msg = 'Unknown IP version "%s".' % data['ip_version']
            raise DataError(msg)

    def _replace_policy_time(self, policy):
        """
        Replaces the time terms in the policy string with unix seconds terms.
        """
        policy_list = str.split(policy)
        self._convert_time(policy_list, 'start')
        self._convert_time(policy_list, 'end')
        converted_policy = ''
        for item in policy_list:
            converted_policy += ' ' + str(item)
        return converted_policy[1:]

    @staticmethod
    def _convert_time(p_list, tag):
        """
        Converts and replaces the time terms after 'tag' in 'p_list' with the
        equivalents in unix seconds.
        """
        try:
            i = p_list.index(tag)
        except:
            return
        try:
            if p_list[i + 1] in COMP:
                next = i + 2
            else:
                next = i + 1
            time = datetime(year=int(p_list.pop(next)),  # year
                            month=int(p_list.pop(next)),  # month
                            day=int(p_list.pop(next)),  # day
                            hour=int(p_list.pop(next)),  # hour
                            minute=int(p_list.pop(next)),  # minute
                            second=int(p_list.pop(next)))  # second
        except (KeyError, ValueError):
            msg = 'Invalid syntax of %s time in policy string.' % tag
            raise ConfigError(msg)
        p_list.insert(next, int((time - TIME_ZERO).total_seconds()))


class Wrapper(object):
    """
    Class that wraps the mixmod dict to an object that returns the mixmod
    fields as its attributes.
    """

    def __init__(self, data):
        self.data = data

    def __getattribute__(self, name):
        data = object.__getattribute__(self, 'data')
        try:
            if name == 'srcip' or name == 'dstip':
                if data['ip_version'] == 4:
                    return ip2int(str(data[MAP[name][0]]))
                if data['ip_version'] == 6:
                    return ip2int(str(data[MAP[name][1]]))
            elif name == 'start':
                return int((data['utc_first_packet'] - TIME_ZERO).total_seconds())
            elif name == 'end':
                return int((data['utc_last_packet'] - TIME_ZERO).total_seconds())
            else:
                return data[MAP[name]]
        except KeyError as e:
            raise DataError('Field does not exist in mixmod record: ' + str(e))
