# jsonimporter.py
#
# Copyright (c) 2015 DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from mixmod import importer
from mixmod.errors import DataError

from datetime import datetime
import json
import netaddr
import unicodedata

EPOCH = datetime(1970, 1, 1)

class JSONImporter(importer.FileImporter):
    CONFIG = {}

    def __init__(self, fobj, **conf):
        super(JSONImporter, self).__init__(fobj, **conf)
        self.iterator = self._parse()

    def get(self):
        try:
            return next(self.iterator)
        except StopIteration:
            return None

    def _parse(self):
        try:
            jsondata = json.loads(self.fobj.getvalue().decode(
                'utf-8').replace("'", '"'))
        except TypeError as e:
            msg = 'unable to parse input data: {}'.format(e)
            raise DataError(msg)

        if isinstance(jsondata, dict):
            mm_dict = self._parse_single(jsondata)
            yield mm_dict

        elif isinstance(jsondata, list):
            for report in jsondata:
                yield self._parse_single(report)

    def _parse_single(self, raw_data):
        for field in raw_data.keys():
            if field in type_conversions.keys():
                try:
                    raw_data[field] = type_conversions[field](raw_data[field])
                except:
                    pass
            else:
                try:
                    raw_data[field] = int(raw_data[field])
                except ValueError:
                    if raw_data[field] == '':
                        raw_data[field] = 0
            if isinstance(raw_data[field], str):
                normalized = unicodedata.normalize("NFKD", raw_data[field])
                raw_data[field] = normalized.encode("ascii", "ignore")
        return raw_data

def to_datetime(dt):
    if dt is None:
        return EPOCH
    if isinstance(dt, str):
        # 2015-07-28 12:14:13.977000
        #  if the microseconds part is 0, it seems to be usually omitted.
        try:
            return datetime.strptime(dt, '%Y-%m-%d %H:%M:%S.%f')
        except ValueError:
            try:
                datetime.strptime(dt, '%Y-%m-%d %H:%M:%S')
            except:
                return EPOCH
    else:
        raise DataError('str expected, but found {}'.format(type(dt)))

def to_ipaddr(ip):
    if ip is None:
        return None
    if isinstance(ip, str):
        return netaddr.IPAddress(ip)
    else:
        raise DataError('str expected, but found {}'.format(type(ip)))

type_conversions = {
    'c2_ip_v4': to_ipaddr,
    'c2_ip_v6': to_ipaddr,
    'dst_ip_v4': to_ipaddr,
    'dst_ip_v6': to_ipaddr,
    'reported_at': to_datetime,
    'src_ip_v4': to_ipaddr,
    'src_ip_v6': to_ipaddr,
    'timestamp': to_datetime,
    'utc_first_packet': to_datetime,
    'utc_last_packet': to_datetime,
}
