# nfcapexporter.py
#
# Copyright (c) 2015  DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from datetime import datetime
import struct
import io

from mixmod import exporter
from mixmod import nfcap
from mixmod.errors import DataError

EPOCH = datetime(1970, 1, 1)
NFCAP_MAGIC = 0xA50C
NFCAP_MAX_BLOCKSIZE = 1000000
NFCAP_VERSION = 1
NFCAP_BLOCK_ID = 2
NFCAP_BLOCK_FLAG_COMPAT = 0
V1_EXTENSION_MAP_ID = 2
COMMON_RECORD_ID = 1
BIT16 = (2 ** 16) - 1
BIT32 = (2 ** 32) - 1

FLAG_IPV6_ADDR          = (1 << 0)
FLAG_PKG_64             = (1 << 1)
FLAG_BYTES_64           = (1 << 2)
FLAG_NEXT_HOP_64        = (1 << 3)
FLAG_NEXT_HOP_BGP_64    = (1 << 4)
FLAG_EXPORTER_IP_64     = (1 << 5)
FLAG_EVENT              = (1 << 6)
FLAG_SAMPLED_FLOW       = (1 << 7)

INTERFACES_4_5 = ['snmp_index_in', 'snmp_index_out']
AS_RECORDS_6_7 = ['src_as', 'dst_as']
MULTIPLE_FIELDS_8 = ['tos', 'dir', 'srcmask', 'dstmask']
VLAN_RECORDS_13 = ['srcvlan', 'dstvlan']
MAC_ADDRESSES_20 = ['in_src_mac', 'out_dst_mac']
MORE_MAC_ADDRESSES_21 = ['in_dst_mac', 'out_dst_mac']
MPLS_LABELS_22 = ['mpls_label_2', 'mpls_label_1', 'mpls_label_4',
                  'mpls_label_3', 'mpls_label_6', 'mpls_label_5',
                  'mpls_label_8', 'mpls_label_7', 'mpls_label_10',
                  'mpls_label_9']
ROUTER_SOURCE_ID_25 = ['fill_e25', 'engine_type', 'engine_id']


class NFCapExporter(exporter.Exporter):
    CONFIG = {
        "standard_direction": {
            "type": "int",
            "default": "0",
            "ident": ""
        }
    }

    def __init__(self, fobj,
                 standard_direction=0,  # 0 (ingress) or 1 (egress)
                 ident="",
                 **conf):
        super(NFCapExporter, self).__init__(**conf)
        self.fobj = fobj
        self.standard_direction = standard_direction
        self.outfile = NFCapFile(self.fobj, standard_direction, ident)

    def put(self, data):
        self.outfile.add_record(data)

    def close(self):
        self.outfile.close()


class NFCapFile(object):
    def __init__(self, fobj, standard_direction, ident):
        self.fobj = fobj
        self.ident = ident
        self.standard_direction = standard_direction
        self.current_map_id = 42
        self.extension_maps = {}
        self.file_header = NFCapFileHeader(self.fobj, ident)
        self.file_header.write()
        self.stat_record = NFCapStatRecord(self.fobj)
        self.stat_record.write()
        self.current_block = None

    def add_record(self, data):
        self.stat_record.count_data_record(data)
        if self.current_block is None:
            self.current_block = NFCapDataBlock(self.fobj, self.standard_direction, self)
            self.file_header.count_data_block(data)
            self.current_block.write()
        self.current_block.add_record(data)
        if self.current_block.size > NFCAP_MAX_BLOCKSIZE:
            self.current_block.close()
            self.current_block = None

    def get_extension_map_id(self, used_extensions, record_flags, current_block):
        used_ext_sorted = ','.join(str(x) for x in sorted(used_extensions))
        if used_ext_sorted in self.extension_maps:
            return self.extension_maps[used_ext_sorted]
        else:
            self.current_map_id += 1
            extension_map = NFCapExtensionMap(self.fobj,
                                              self.current_map_id,
                                              used_extensions,
                                              record_flags,
                                              current_block)
            extension_map.write()
            self.extension_maps[used_ext_sorted] = extension_map.id
            return extension_map.id

    def close(self):
        if self.current_block is not None:
            self.current_block.close()
        self.file_header.close()
        self.stat_record.close()


class NFCapFileSection(object):
    def __init__(self, fobj):
        self.fobj = fobj
        self.start_position = self.fobj.tell()
        self.old_pos = self.fobj.tell()

    def write(self):
        raise NotImplementedError

    def close(self):
        old_pos = self.fobj.tell()
        self.fobj.seek(self.start_position)
        self.write()
        self.fobj.seek(old_pos)


class NFCapFileHeader(NFCapFileSection):
    def __init__(self, fobj, ident):
        super(NFCapFileHeader, self).__init__(fobj)
        self.flags = 0
        self.numblocks = 0
        self.ident = ident

    def write(self):
        data = nfcap.nfcap_header.pack(
            NFCAP_MAGIC,
            NFCAP_VERSION,
            self.flags,
            self.numblocks,
            self.ident.encode('utf-8')
        )
        self.fobj.write(data)

    def count_data_block(self, data):
        self.numblocks += 1


class NFCapStatRecord(NFCapFileSection):
    def __init__(self, fobj):
        super(NFCapStatRecord, self).__init__(fobj)
        self.numflows = 0
        self.numbytes = 0
        self.numpackets = 0
        self.numflows_tcp = 0
        self.numflows_udp = 0
        self.numflows_icmp = 0
        self.numflows_other = 0
        self.numbytes_tcp = 0
        self.numbytes_udp = 0
        self.numbytes_icmp = 0
        self.numbytes_other = 0
        self.numpackets_tcp = 0
        self.numpackets_udp = 0
        self.numpackets_icmp = 0
        self.numpackets_other = 0
        self.first_seen = (datetime.now() - EPOCH).total_seconds()
        self.last_seen = self.first_seen
        self.sequence_failure = 0

    def write(self):
        data = nfcap.nfcap_stats.pack(
            self.numflows,
            self.numbytes,
            self.numpackets,

            self.numflows_tcp,
            self.numflows_udp,
            self.numflows_icmp,
            self.numflows_other,

            self.numbytes_tcp,
            self.numbytes_udp,
            self.numbytes_icmp,
            self.numbytes_other,

            self.numpackets_tcp,
            self.numpackets_udp,
            self.numpackets_icmp,
            self.numpackets_other,

            int(self.first_seen),
            int(self.last_seen),
            int((self.first_seen - int(self.first_seen)) * 1000),
            int((self.last_seen - int(self.last_seen)) * 1000),
            self.sequence_failure
        )
        self.fobj.write(data)

    def count_data_record(self, data):
        ip_proto = data.get('ip_protocol_number', 0)
        packet_count = data.get('packet_count', 0)
        byte_count = data.get('byte_count', 0)
        self.numflows += 1
        self.numbytes += byte_count
        self.numpackets += packet_count
        if ip_proto == 1:
            self.numflows_icmp += 1
            self.numpackets_icmp += packet_count
            self.numbytes_icmp += byte_count
        elif ip_proto == 6:
            self.numflows_tcp += 1
            self.numpackets_tcp += packet_count
            self.numbytes_tcp += byte_count
        elif ip_proto == 17:
            self.numflows_udp += 1
            self.numpackets_udp += packet_count
            self.numbytes_udp += byte_count
        else:
            self.numflows_other += 1
            self.numpackets_other += packet_count
            self.numbytes_other += byte_count

        try:
            first_ts = (data.get('utc_first_packet', EPOCH) - EPOCH).total_seconds()
        except TypeError as e:
            first_ts = 0
            # raise TypeError('{}\n\n{}'.format(data, e))
        try:
            last_ts = (data.get('utc_last_packet', EPOCH) - EPOCH).total_seconds()
        except TypeError as e:
            last_ts = 0
            # raise TypeError('{}\n\n{}'.format(data, e))
        if (first_ts is not None) and (self.first_seen > first_ts):
            self.first_seen = first_ts
        if (last_ts is not None) and (self.last_seen < last_ts):
            self.last_seen = last_ts

        self.sequence_failure += data.get('sequence_failure', 0)


class NFCapDataBlock(NFCapFileSection):
    def __init__(self, fobj, standard_direction, encapsulating_file):
        super(NFCapDataBlock, self).__init__(fobj)
        self.standard_direction = standard_direction
        self.encapsulating_file = encapsulating_file
        self.numrecords = 0
        self.size = 0
        self.id = NFCAP_BLOCK_ID
        self.blockflags = NFCAP_BLOCK_FLAG_COMPAT

    def write(self):
        self.start_position = self.fobj.tell()
        data = nfcap.nfcap_blockheader.pack(
            self.numrecords,
            self.size,
            self.id,
            self.blockflags
        )
        self.fobj.write(data)

    def add_record(self, record):
        data_record = NFCapDataRecord(self.fobj,
                                      record,
                                      self.standard_direction,
                                      self.encapsulating_file,
                                      self)
        if data_record.ip_version is None:
            return
        self.numrecords += 1
        data_record.write()
        self.size += data_record.recordsize()


class NFCapExtensionMap(NFCapFileSection):
    def __init__(self, fobj, ext_map_id, used_extensions, record_flags, encapsulating_block):
        super(NFCapExtensionMap, self).__init__(fobj)
        self.ext_map_id = ext_map_id
        self.buf = io.BytesIO()
        self.numrecords = 0
        self.numrecords_without_padding = 0
        self.mapsize = 0
        self.extension_size = 0
        self.used_extensions = used_extensions
        self.encapsulating_block = encapsulating_block

        for extension_id in used_extensions:
            # Note that the mandatory extensions (1, 2, 3) do not appear in extension maps
            # nor size calculations.
            # if extension_id == 1:
            #     if record_flags & FLAG_IPV6_ADDR:
            #         self.extension_size += 32
            #     else:
            #         self.extension_size += 8
            # elif extension_id == 2:
            #     if record_flags & FLAG_PKG_64:
            #         self.extension_size += 8
            #     else:
            #         self.extension_size += 4
            # elif extension_id == 3:
            #     if record_flags & FLAG_BYTES_64:
            #         self.extension_size += 8
            #     else:
            #         self.extension_size += 4
            if extension_id not in [1, 2, 3]:
                self.extension_size += _get_extension_struct_size(extension_id)
                self.numrecords += 1
                data = nfcap.nfcap_record_v2_extension_map_entry.pack(
                    extension_id
                )
                self.buf.write(data)
        self.numrecords_without_padding = self.numrecords
        self.numrecords += 1
        data = nfcap.nfcap_record_v2_extension_map_entry.pack(
            0
        )
        self.buf.write(data)
        # extension map must be 32-bit aligned:
        if self.numrecords % 2 != 0:
            data = nfcap.nfcap_record_v2_extension_map_entry.pack(
                0
            )
            self.buf.write(data)
            self.numrecords += 1
        self.mapsize = self.numrecords * 2

    @property
    def id(self):
        return self.ext_map_id

    def write(self):
        data = nfcap.nfcap_extension_map_record.pack(
            V1_EXTENSION_MAP_ID,
            self.blocksize,
            self.ext_map_id,
            self.mapsize,
        )
        self.fobj.write(data)
        self.fobj.write(self.buf.getvalue())
        self.encapsulating_block.size += self.blocksize
        self.encapsulating_block.numrecords += 1

    @property
    def blocksize(self):
        return nfcap.nfcap_extension_map_record.size + self.mapsize


class NFCapDataRecord(NFCapFileSection):
    def __init__(self, fobj, data, standard_direction, encapsulating_file, encapsulating_block):
        super(NFCapDataRecord, self).__init__(fobj)

        self.data_dict = data
        self.direction = data.get('direction', standard_direction)
        self.file_object = encapsulating_file
        self.block_object = encapsulating_block

        self.ext_map_id = 0
        self.nfcap_flags = 0
        self.used_extension_ids = []

        self.mark = data.get('mark', 0)

        # Values that must be written in the "common record" or a mandatory extension
        # are read into instance variables here.
        # However, the data dictionary itself is accessed by the methods creating
        # optional extensions, so they can easily determine whether or not to write
        # to self.buf.

        if 'ip_version' in data:
            self.ip_version = data.get('ip_version')
        else:
            if 'src_ip_v4' in data and 'dst_ip_v4' in data:
                self.ip_version = 4
            elif 'src_ip_v6' in data and 'dst_ip_v6' in data:
                self.ip_version = 6
            else:
                # Cannot find any IP version information (nor src/dst IP addresses), abort this flow.
                self.ip_version = None
                return

        if 'utc_first_packet' in data and data['utc_first_packet'] is not None:
            first_secs = int((data['utc_first_packet'] - EPOCH).total_seconds())
            self.first = int(first_secs)
            self.msec_first = (first_secs - self.first) * 1000
        else:
            self.first = 0
            self.msec_first = 0
        if 'utc_last_packet' in data and data['utc_last_packet'] is not None:
            last_secs = (data['utc_last_packet'] - EPOCH).total_seconds()
            self.last = int(last_secs)
            self.msec_last = int((last_secs - self.last) * 1000)
        else:
            self.last = 0
            self.msec_last = 0

        if self.ip_version == 4:
            self.src_ip_v4 = data.get('src_ip_v4', None)
            self.dst_ip_v4 = data.get('dst_ip_v4', None)
        elif self.ip_version == 6:
            self.nfcap_flags |= FLAG_IPV6_ADDR
            self.src_ip_v6 = data.get('src_ip_v6', None)
            self.dst_ip_v6 = data.get('dst_ip_v6', None)
        else:
            raise DataError('no ip address in flow')

        self.next_hop_ip_version = data.get('next_hop_ip_version', None)
        if self.next_hop_ip_version == 6:
            self.nfcap_flags |= FLAG_NEXT_HOP_64
        self.next_hop_bgp_ip_version = data.get('next_bgp_ip_version', None)
        if self.next_hop_bgp_ip_version == 6:
            self.nfcap_flags |= FLAG_NEXT_HOP_BGP_64

        self.exporter_id = data.get('organisation_id', 0)

        self.packet_count = data.get('packet_count', 0)
        if self.packet_count > BIT32:
            self.nfcap_flags |= FLAG_PKG_64
        self.byte_count = data.get('byte_count', 0)
        if self.byte_count > BIT32:
            self.nfcap_flags |= FLAG_BYTES_64

        self.tcp_flags = data.get('tcp_flags', 0)
        self.proto = data.get('ip_protocol_number', 0)
        self.tos = data.get('type_of_service', 0)
        self.in_if = data.get('snmp_index_in', None)
        self.out_if = data.get('snmp_index_out', None)
        self.src_port = data.get('src_port', 0)
        self.dst_port = data.get('dst_port', 0)
        self.src_as = data.get('src_as', None)
        self.dst_as = data.get('dst_as', None)

        self.fwd_status = data.get('fwd_status', 0)

        self.buf = io.BytesIO()

    def write(self):
        #
        # The "generate*" and "add*" methods only write to the BytesIO buffer in self.buf,
        # so that the total size of the record can be determined before actually writing to the
        # output file object. "write*" methods write to the output file.

        # mandatory extensions: IP addresses, packet count, byte count
        self.add_ext_IP()
        self.add_ext_packet()
        self.add_ext_bytes()

        # optional extensions: these write to self.buf only if they find the data
        # for which they are responsible in self.datadict.
        self.add_optional_size_dependent_pair(
            ['snmp_index_in', 'snmp_index_out'],
            4, 5)
        self.add_optional_size_dependent_pair(
            ['src_as', 'dst_as'],
            6, 7
        )
        self.add_ex_multiple()
        self.add_optional_ipaddress(
            'next_hop_ip',
            9, 10
        )
        self.add_optional_ipaddress(
            'next_bgp_ip',
            11, 12
        )
        self.add_ex_vlan()
        self.add_optional_size_dependent_single(
            'numoutpackets',
            14, 15
        )
        self.add_optional_size_dependent_single(
            'numoutbytes',
            16, 17
        )
        self.add_optional_size_dependent_single(
            'aggrflows',
            18, 19
        )
        self.add_optional_macaddress_pair(
            ['in_src_mac', 'out_dst_mac'],
            20
        )
        self.add_optional_macaddress_pair(
            ['in_dst_mac', 'out_dst_mac'],
            21
        )
        self.add_ex_mpls()
        self.add_optional_ipaddress(
            'router_ip',
            23, 24
        )
        self.add_ex_router_source()
        self.ext_map_id = self.file_object.get_extension_map_id(self.used_extension_ids,
            self.nfcap_flags, self.block_object)
        self.write_v16_record_header()
        self.write_v16_common_record()
        self.fobj.write(self.buf.getvalue())

    def recordsize(self):
        return nfcap.nfcap_record_v2_header.size + nfcap.nfcap_record_v2_common.size + \
               len(self.buf.getvalue())

    def write_v16_record_header(self):
        data = nfcap.nfcap_record_v2_header.pack(
            1,  # record_type = 1: CommonRecordV0Type
            self.recordsize()
        )
        self.fobj.write(data)

    def write_v16_common_record(self):
        data = nfcap.nfcap_record_v2_common.pack(
            self.nfcap_flags,
            self.exporter_id,
            self.ext_map_id,
            self.msec_first, self.msec_last,
            self.first, self.last,
            self.fwd_status,
            self.tcp_flags,
            self.proto,
            self.tos,
            self.src_port,
            self.dst_port
        )
        self.fobj.write(data)

    def add_ext_IP(self):
        if self.ip_version == 4:
            data = nfcap.ipv4x2.pack(
                self.src_ip_v4.value,
                self.dst_ip_v4.value
            )
            self.buf.write(data)
        elif self.ip_version == 6:
            srcip = self.src_ip_v6.value
            dstip = self.dst_ip_v6.value
            data = nfcap.ipv6x2.pack(
                srcip // 2 ** 64,
                srcip % 2 ** 64,
                dstip // 2 ** 64,
                dstip % 2 ** 64,
            )
            self.buf.write(data)
        self.used_extension_ids.append(1)

    def add_ext_packet(self):
        if self.nfcap_flags & FLAG_PKG_64:
            data = nfcap.u64.pack(
                self.packet_count
            )
            self.buf.write(data)
        else:
            data = nfcap.u32.pack(
                self.packet_count
            )
            self.buf.write(data)
        self.used_extension_ids.append(2)

    def add_ext_bytes(self):
        if self.nfcap_flags & FLAG_BYTES_64:
            data = nfcap.u64.pack(
                self.byte_count
            )
            self.buf.write(data)
        else:
            data = nfcap.u32.pack(
                self.byte_count
            )
            self.buf.write(data)
        self.used_extension_ids.append(3)

    def add_optional_size_dependent_single(self, field_name, opt16bit, opt32bit):
        value = self.data_dict.get(field_name, None)
        if value is None:
            return

        if value < BIT16:
            use_option = opt16bit
        else:
            use_option = opt32bit
        if use_option is None:
            return

        data = _get_extension_struct_def(use_option).pack(
            value
        )
        self.buf.write(data)
        self.used_extension_ids.append(use_option)

    def add_optional_size_dependent_pair(self, field_list, opt16bit, opt32bit):
        if not all((field in self.data_dict and self.data_dict[field] is not None)
                for field in field_list):
            return
        if all(self.data_dict[field] < BIT16 for field in field_list):
            use_option = opt16bit
        else:
            use_option = opt32bit
        if use_option is None:
            return
        data = _get_extension_struct_def(use_option).pack(
            self.data_dict[field_list[0]],
            self.data_dict[field_list[1]],
        )
        self.buf.write(data)
        self.used_extension_ids.append(use_option)

    def add_optional_ipaddress(self, field_name, optv4, optv6):
        ipaddr = self.data_dict.get(field_name, None)
        if ipaddr is None:
            return

        ipaddr = ipaddr.value

        if self.ip_version == 4:
            if optv4 is None:
                return
            data = _get_extension_struct_def(optv4).pack(
                ipaddr
            )
            self.buf.write(data)
            self.used_extension_ids.append(optv4)
        elif self.ip_version == 6:
            if optv6 is None:
                return
            data = _get_extension_struct_def(optv6).pack(
                ipaddr // 2 ** 64,
                ipaddr % 2 ** 64
            )
            self.buf.write(data)
            self.used_extension_ids.append(optv6)

    def add_optional_macaddress_pair(self, field_list, opt_id):
        if not all((field in self.data_dict and self.data_dict[field] is not None)
                for field in field_list):
            return
        data = _get_extension_struct_def(opt_id).pack(
            self.data_dict[field_list[0]],
            self.data_dict[field_list[1]],
        )
        self.buf.write(data)
        self.used_extension_ids.append(opt_id)

    def add_ex_multiple(self):
        fields = ['tos', 'dir', 'srcmask', 'dstmask']
        if not all((field in self.data_dict and self.data_dict[field] is not None)
                   for field in fields):
            return
        data = _get_extension_struct_def(8).pack(
            self.data_dict['tos'],
            self.data_dict['dir'],
            self.data_dict['srcmask'],
            self.data_dict['dstmask']
        )
        self.buf.write(data)
        self.used_extension_ids.append(8)

    def add_ex_vlan(self):
        fields = ['srcvlan', 'dstvlan']
        if not all((field in  self.data_dict and self.data_dict[field] is not None)
                   for field in fields):
            return
        data = _get_extension_struct_def(13).pack(
            self.data_dict['srcvlan'],
            self.data_dict['dstvlan']
        )
        self.buf.write(data)
        self.used_extension_ids.append(13)

    def add_ex_mpls(self):
        fields = ['mpls_label_2', 'mpls_label_1', 'mpls_label_4',
                  'mpls_label_3', 'mpls_label_6', 'mpls_label_5',
                  'mpls_label_8', 'mpls_label_7', 'mpls_label_10',
                  'mpls_label_9']
        if not all((field in self.data_dict and self.data_dict[field] is not None)
                for field in fields):
            return
        data = _get_extension_struct_def(22).pack(
            self.data_dict['mpls_label_2'],
            self.data_dict['mpls_label_1'],
            self.data_dict['mpls_label_2'],
            self.data_dict['mpls_label_3'],
            self.data_dict['mpls_label_6'],
            self.data_dict['mpls_label_5'],
            self.data_dict['mpls_label_8'],
            self.data_dict['mpls_label_7'],
            self.data_dict['mpls_label_10'],
            self.data_dict['mpls_label_9'],
        )
        self.buf.write(data)
        self.used_extension_ids.append(22)

    def add_ex_router_source(self):
        fields = ['engine_type', 'engine_id']
        if not all((field in self.data_dict and self.data_dict[field] is not None)
                for field in fields):
            return
        data = _get_extension_struct_def(25).pack(
            0,
            self.data_dict['engine_type'],
            self.data_dict['engine_id']
        )
        self.buf.write(data)
        self.used_extension_ids.append(25)

def _get_extension_struct_def(extension_id):
    return struct.Struct('<' + nfcap.EXTENSION_DESCRIPTIONS[extension_id][0])

def _get_extension_struct_size(extension_id):
    return _get_extension_struct_def(extension_id).size

'''
File Format Discription

+------------+--------------+--------------------+-----+---------------------+
| FileHeader | Stats Record | Datablock Header 1 | ... | Datablock Header n  |
+------------+--------------+--------------------+-----+---------------------+

+--------------------+----------+-----+----------+--------------------+-----+
| Datablock Header 1 | Record 1 | ... | Record n | Datablock Header n | ... |
+--------------------+----------+-----+----------+--------------------+-----+

'''

nfcap_fileheader = struct.Struct("<"
                             "H"  # u16 magic, NFCAP_MAGIC
                             "H"  # u16 version, should be 1 (v1.5 format)
                             "I"  # u32 flags, bitmask for see NFCAP_FLAG_*
                             "I"  # u32 numblocks, count of blocks in the file
                             "128s"  # ident, free-form string
                            )

nfcap_statsrecord = struct.Struct("<"
                            "Q"  # u64 num flows
                            "Q"  # u64 num bytes
                            "Q"  # u64 num packets
                            "Q"  # u64 num flows tcp
                            "Q"  # u64 num flows udp
                            "Q"  # u64 num flows icmp
                            "Q"  # u64 num flows other
                            "Q"  # u64 num bytes tcp
                            "Q"  # u64 num bytes udp
                            "Q"  # u64 num bytes icmp
                            "Q"  # u64 num bytes other
                            "Q"  # u64 num packets tcp
                            "Q"  # u64 num packets udp
                            "Q"  # u64 num packets icmp
                            "Q"  # u64 num packets other
                            "I"  # u32 first seen, unix time stamp
                            "I"  # u32 last seen, unix time stamp
                            "H"  # u16 first seen milliseconds
                            "H"  # u16 last seen milliseconds
                            "I"  # u32 sequence failures, no clue what this is
                            )

nfcap_datablockheader = struct.Struct("<"
                                      "I"   # u32 numrecords, number of records
                                      "I"   # u32 size of this datablock
                                      "H"   # u16 id, data block type
                                      "H"   # u16 pad, padding
                                      )

nfcap_record_v2_header = struct.Struct("<"
                                       "H"  # u16 record id
                                       "H"  # u16 record size
                                       )

nfcap_extension_map_basic = struct.Struct("<"
                                    "H" # u16 map id
                                    "H" # u16 size
                                    "H" # u16 ref map id
                                    "H" # u16 ext size
                                    )

nfcap_extension_map_field = struct.Struct("<"
                                          "H"
                                         )

nfcap_extension_map_4_6 = struct.Struct("<"
                                       "H" # u16 input
                                       "H" # u16 output
                                       "H" # u16 srcas
                                       "H" # u16 dstas
                                       )

nfcap_extension_map_5_7 = struct.Struct("<"
                                       "I" # u32 input
                                       "I" # u32 output
                                       "I" # u32 srcas
                                       "I" # u32 dstas
                                       )

nfcap_record_v2_common = struct.Struct("<"
                                       "H"  # u16 id
                                       "H"  # u16 size
                                       "B"  # u8 flags
                                       "B"  # u8 tag
                                       "H"  # u16 extension map
                                       "H"  # u16 msec_first
                                       "H"  # u16 msec_last
                                       "I"  # u32 first
                                       "I"  # u32 last
                                       "BBBB"  # u8 fwd_status, tcp, prot, tos
                                       "H"  # u16 srcport
                                       "H"  # u16 dstport
                                        )

nfcap_packets_bytes_32 = struct.Struct("<"
                          "I" # u32 number of packets or bytes
                         )

nfcap_packets_bytes_64 = struct.Struct("<"
                          "Q" # u64 packets or bytes
                         )

nfcap_v4block = struct.Struct("<"
                              "I" # u32 source ipv4 address
                              "I" # u32 destination ipv4 address
                              )

nfcap_v6block =  struct.Struct("<"
                                "Q" # u64 source ipv6 address ip1a
                                "Q" # u64 source ipb4 address ip1b
                                "Q" # u64 destination ipv6 address ip2a
                                "Q" # u64 destination ipv6 address ip2b
                                )
