# pcapimporter.py
#
# Copyright (c) 2015 DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import socket
import struct
import sys
from datetime import datetime, timedelta
from math import floor
from struct import unpack_from

import dpkt
import netaddr

from mixmod import errors, importer

# definition of pcap global header field order and length
GLOBAL_HEADER_FORMAT_STRING = (
    'I'     # uint32 magic_number / magic number
    'H'     # uint16 version_major / major version number
    'H'     # uint16 version_minor / minor version number
    'i'     # int32  thiszone / GMT to local correction
    'I'     # uint32 sigfigs / accuracy of timestamps
    'I'     # uint32 snaplen / max length of captured packets, in octets
    'I'     # uint32 network / data link type
)
GLOBAL_HEADER_LENGTH = 24

# definition of pcap packet header field order and length
PACKET_HEADER_FORMAT_STRING = (
    'I'     # guint32 ts_sec;         /* timestamp seconds */
    'I'     # guint32 ts_usec;        /* timestamp microseconds */
    'I'     # guint32 incl_len;       /* number of octets of packet saved in file */
    'I'     # guint32 orig_len;       /* actual length of packet */
)
PACKET_HEADER_LENGTH = 16

# additional fields that may be set via configuration
ADDITIONAL_FIELDS = ["tlp", "src_mode", "dst_mode", "attachment", "impact",
                     "enviroment", "location", "ids_signature_id", "ids_origin",
                     "fwz_analyzer_id", "fwz_sector", "fwz_organisation_id",
                     "organisation_id", "attachment_type"]

SUPPORTED_PROTOCOLS = [dpkt.ip.IP_PROTO_TCP, dpkt.ip.IP_PROTO_UDP, dpkt.ip.IP_PROTO_ICMP, dpkt.ip.IP_PROTO_ICMP6]


class PCapImporter(importer.FileImporter):
    """Import mixmod data from pcap files."""

    DIRECTION_INGRESS = 0
    DIRECTION_EGRESS = 1

    CONFIG = {
        "standard_direction": {
            "type": "int",
            "default": DIRECTION_INGRESS
        }
    }

    def __init__(self, fobj, standard_direction=DIRECTION_INGRESS, **conf):
        super(PCapImporter, self).__init__(fobj, **conf)
        self.mixmoddata = dict()
        self.offset = 0
        self.direction = list()
        self.standard_direction = int(standard_direction)
        self.nano = False

        for key in conf.keys():
            if key.startswith("src_ingress"):
                self.direction.append({'use': "src",
                                       'netmask': conf[key],
                                       'direction': self.DIRECTION_INGRESS})
                del conf[key]
            elif key.startswith("dst_ingress"):
                self.direction.append({'use': "dst",
                                       'netmask': conf[key],
                                       'direction': self.DIRECTION_INGRESS})
                del conf[key]
            elif key.startswith("src_egress"):
                self.direction.append({'use': "src",
                                       'netmask': conf[key],
                                       'direction': self.DIRECTION_EGRESS})
                del conf[key]
            elif key.startswith("dst_egress"):
                self.direction.append({'use': "dst",
                                       'netmask': conf[key],
                                       'direction': self.DIRECTION_EGRESS})
                del conf[key]

        self.additional_fields = dict()
        for field in ADDITIONAL_FIELDS:
            if field in conf:
                self.additional_fields[field] = conf[field]
                del conf[field]

        self.rawdata = fobj.read()
        self.iterator = self._parse()

    def get(self):
        try:
            return next(self.iterator)
        except StopIteration:
            return None

    def _parse(self):
        """
        Parses the data in self.rawdata by using helper functions and generates
        mixmoddata dictionaries.

        This method will parse all data that is available in self.rawdata, aggregate it into
        flows which will be saved in the aforementioned mixmoddata dictionaries.
        """
        self._parse_global_header()
        while True:
            header = self._parse_header()
            if header is False:
                break

            # read packet
            eth = dpkt.ethernet.Ethernet(self.rawdata[self.offset:(self.offset+header['incl_len'])])
            self.offset += header['incl_len']

            if not (eth.type == dpkt.ethernet.ETH_TYPE_IP or eth.type == dpkt.ethernet.ETH_TYPE_IP6):
                continue  # ignore everything but IP/IPv6

            if eth.data.p not in SUPPORTED_PROTOCOLS:
                continue  # ignore

            src_ip = eth.data.src
            dst_ip = eth.data.dst
            # create hash from srcip and dstip and port
            # to identify flows
            self.mixmoddata = {
                'tcp_flags': 0,
                'ip_protocol_number': eth.data.p,
                'utc_first_packet': datetime.utcfromtimestamp(header['ts_sec']) + timedelta(milliseconds=header['ts_usec']),
                'utc_last_packet': datetime.utcfromtimestamp(header['ts_sec']) + timedelta(milliseconds=header['ts_usec']),
                'packet_count': 1,
                'byte_count': header['orig_len'],  # TODO: Check if orig_len or incl_len
                'original_format': 'pcap',
            }

            if eth.type == dpkt.ethernet.ETH_TYPE_IP6:
                self.mixmoddata['src_ip_v6'] = netaddr.IPAddress(socket.inet_ntop(socket.AF_INET6, src_ip))
                self.mixmoddata['dst_ip_v6'] = netaddr.IPAddress(socket.inet_ntop(socket.AF_INET6, dst_ip))
                self.mixmoddata['ip_version'] = 6
                self.mixmoddata['direction'] = self._get_direction({'src': self.mixmoddata['src_ip_v6'], 'dst': self.mixmoddata['dst_ip_v6']})
            elif eth.type == dpkt.ethernet.ETH_TYPE_IP:
                self.mixmoddata['src_ip_v4'] = netaddr.IPAddress(socket.inet_ntop(socket.AF_INET, src_ip))
                self.mixmoddata['dst_ip_v4'] = netaddr.IPAddress(socket.inet_ntop(socket.AF_INET, dst_ip))
                self.mixmoddata['ip_version'] = 4
                self.mixmoddata['direction'] = self._get_direction({'src': self.mixmoddata['src_ip_v4'], 'dst': self.mixmoddata['dst_ip_v4']})

            if eth.data.p == dpkt.ip.IP_PROTO_TCP or eth.data.p == dpkt.ip.IP_PROTO_UDP:
                self.mixmoddata['src_port'] = eth.data.data.sport
                self.mixmoddata['dst_port'] = eth.data.data.dport
            if eth.data.p == dpkt.ip.IP_PROTO_TCP:
                self.mixmoddata['tcp_flags'] = eth.data.data.flags
                if eth.type == dpkt.ethernet.ETH_TYPE_IP:
                    self.mixmoddata['type_of_service'] = eth.data.tos
                elif eth.type == dpkt.ethernet.ETH_TYPE_IP6:
                    self.mixmoddata['type_of_service'] = eth.data.fc

            if eth.data.p == dpkt.ip.IP_PROTO_ICMP:
                self.mixmoddata['icmp_type_ipv4'] = eth.data.data.type
                self.mixmoddata['icmp_code_ipv4'] = eth.data.data.code
            if eth.data.p == dpkt.ip.IP_PROTO_ICMP6:
                self.mixmoddata['icmp_type_ipv6'] = eth.data.data.type
                self.mixmoddata['icmp_code_ipv6'] = eth.data.data.code

            self.mixmoddata.update(self.additional_fields)

            yield self.mixmoddata

    def _parse_global_header(self):
        """
        Parse the pcap global header
        """
        (magic_number,) = unpack_from("I", self.rawdata)
        self.endian = ":"
        if magic_number == 0xa1b2c3d4:
            if sys.byteorder == "little":
                self.endian = "<"
            else:
                self.endian = ">"
        elif magic_number == 0xd4c3b2a1:
            if sys.byteorder == "little":
                self.endian = ">"
            else:
                self.endian = "<"
        elif magic_number == 0xa1b23c4d:
            self.nano = True
            if sys.byteorder == "little":
                self.endian = ">"
            else:
                self.endian = "<"
        elif magic_number == 0x4d3cb2a1:
            self.nano = True
            if sys.byteorder == "little":
                self.endian = ">"
            else:
                self.endian = "<"

        (mn, version_major, version_minor, timezone, sigifs, snaplen, network) = unpack_from(self.endian + GLOBAL_HEADER_FORMAT_STRING, self.rawdata)
        if network != 1:
            raise errors.LinkTypeError("Unsupported link type (%s) in pcap file header" % network)
        self.offset += GLOBAL_HEADER_LENGTH

    def _parse_header(self):
        """
        Parse the pcap packet header
        """
        try:
            (ts_sec, ts_usec, incl_len, orig_len) = unpack_from(self.endian + PACKET_HEADER_FORMAT_STRING, self.rawdata, self.offset)
        except struct.error:
            return False
        self.offset += PACKET_HEADER_LENGTH
        if self.nano:
            ts_usec = floor(ts_usec // 1000)
        return {"ts_sec": ts_sec, "ts_usec": ts_usec, "incl_len": incl_len, "orig_len": orig_len}

    def _get_direction(self, ip):
        """
        Determine direction by checking if src or dst address
        is within a given network (set in configuration).
        """
        for direct in self.direction:
            if netaddr.IPAddress(ip[direct['use']]) in netaddr.IPNetwork(direct['netmask']):
                return direct['direction']
        return self.standard_direction  # default to ingress
