# pseudonymizer.py
#
# Copyright (c) 2014, 2015 DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from collections import deque
import hashlib
import random

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from netaddr import IPAddress

from mixmod.errors import ConfigError, DataError
from mixmod import dataprocessor


class Pseudonymizer(dataprocessor.DataProcessor):
    """This class implements anonymization and pseudonymization for MixModData.

    Anonymization means that data is replaced by random values.

    Pseudonymization means that the same original value is replaced by
    the same pseudonymous value every time it is found to enable
    identity checks, but without the ability to calculate the
    original.

    """

    MODE_PLAIN = 0
    MODE_ANON = 1
    MODE_PSEUDO = 2
    # When one of these fields is updated, also set the assigned mode
    # field.
    MODE_FIELDS = {
        "dst_ip_v4": "dst_mode",
        "dst_ip_v6": "dst_mode",
        "src_ip_v4": "src_mode",
        "src_ip_v6": "src_mode",
        "c2_ip_v4": "c2_mode",
        "c2_ip_v6": "c2_mode",
    }

    IP_BITS = {
        4: 32,
        6: 128
    }

    DEFAULT_MASKLEN = {
        4: 8,
        6: 64
    }

    CONFIG = {}

    def __init__(self, **conf):
        """
        This method initializes a Pseudonymize instance.

        @type  config: ConfigParser.SafeConfigParser
        @param config: A ConfigParser instance into which the MixMod config
                       file has been read

        @type  confsection: string
        @param confsection: An alternate configuration section name to use
        """
        self.field_mapper = {}
        if "@aes_key_set" in conf:
            self.init_aes(conf["@aes_key_set"].encode('ascii'))
        self.domain_seed = conf.get("@domain_seed")

        for option, value in conf.items():
            if option.startswith("@"):
                continue
            field = option.lower()
            action, args, kwargs = _parse_action(value)
            action_func = "action_{}".format(action)
            mapper = getattr(self, action_func, None)
            if mapper is None:
                raise EnvironmentError('Invalid action for {}: {}'
                                       .format(field, action))

            self.field_mapper[field] = (mapper, args, kwargs)

        self.results = deque()

    def init_aes(self, seed):
        """
        Initialize AES key and padding used for pseudonymization.

        Initialize the AES encryption key and padding used in the Crypto-PAn
        pseudonymization from seed. Needs a 32 byte long bytestring.
        """
        if len(seed) != 32:
            raise ConfigError('Need an @aes_key_set of length 32.')

        backend = default_backend()
        # We need to use a weak ECB cipher here to have a consistent
        # pseudonymization. We explicitly want the same input to generate the
        # same output.
        cipher = Cipher(
            algorithms.AES(seed[:16]), modes.ECB(), backend=backend)  # nosec
        self.aes = cipher.encryptor()
        self.padding = self.aes.update(seed[16:])

    def put(self, datadict):
        """The public method for pseudonymizing a data set

        @type  datadict: MixMod dataset dictionary
        @param datadict: Input data set

        @rtype: Mixmod dataset dictionary
        @return: The input dictionary with the configured replacements done
        """
        datadict = datadict.copy()
        for field in datadict.keys():
            mapper = self.field_mapper.get(field)
            if mapper is not None:
                func, args, kwargs = mapper
                func(datadict, field, *args, **kwargs)
        self.results.append(datadict)

    def get(self):
        try:
            return self.results.popleft()
        except IndexError:
            return None

    def update_mode(self, datadict, field, new_mode):
        mode_field = self.MODE_FIELDS.get(field)
        if mode_field is None:
            return
        if datadict.get(mode_field) == self.MODE_ANON:
            # Do not update anonymous to anything else
            return
        datadict[mode_field] = new_mode

    #################
    # Action handlers

    def action_prefix_hash(self, datadict, field):
        '''
        Return an AES-based hash for an IP address, using the configured key.
        This method works prefix-preserving.

        The algorithm used is from Crypto-PAn with the extensions developed by
        Lucent:
        https://www.cc.gatech.edu/computing/Telecomm/projects/cryptopan/
        https://www.cc.gatech.edu/computing/Networking/projects/cryptopan/lucent.shtml

        This version is extended to also support IPv6 addresses.
        '''
        ip = _get_ip_field(datadict, field)

        ip_bits = self.IP_BITS[ip.version]
        # The used endianness here and below is to yield the same results as
        # Crypto-PAn and have consistent results across architectures.
        pad = int.from_bytes(self.padding[:ip_bits//8],
                             byteorder='little')
        result = 0

        for pos in range(ip_bits, 0, -1):
            mask = 2**pos - 1
            input_ = pad ^ (int(ip) & ~mask)
            output = self.aes.update(
                input_.to_bytes(ip_bits//8, byteorder='big')
                + self.padding[ip_bits//8:]
            )
            result |= (
                (int.from_bytes(output, byteorder='big') & 2**127) >> 128-pos
            )

            pad = (
                (pad << 1) & ((2**ip_bits)-1)
                | (pad >> (ip_bits-1))
            )

        datadict[field] = type(ip)(result ^ int(ip))
        self.update_mode(datadict, field, self.MODE_PSEUDO)

    def action_revert_prefix_hash(self, datadict, field):
        '''
        Revert the AES-based hash for an IP address computed by
        action_prefix_hash. This works only if the configured key is the same.
        '''
        ip = _get_ip_field(datadict, field)

        ip_bits = self.IP_BITS[ip.version]
        # Use the same endianness as action_prefix_hash.
        pad = int.from_bytes(self.padding[:ip_bits//8],
                             byteorder='little')
        int_ip = int(ip)

        for pos in range(ip_bits, 0, -1):
            mask = 2**pos - 1
            input_ = pad ^ (int_ip & ~mask)
            output = self.aes.update(
                input_.to_bytes(ip_bits//8, byteorder='big')
                + self.padding[ip_bits//8:]
            )
            int_ip ^= (
                (int.from_bytes(output, byteorder='big') & 2**127) >> 128-pos
            )

            pad = (
                (pad << 1) & ((2**ip_bits)-1)
                | (pad >> (ip_bits-1))
            )

        datadict[field] = type(ip)(int_ip)
        self.update_mode(datadict, field, self.MODE_PLAIN)

    def action_ip_truncate_low(self, datadict, field, masklen=None):
        '''
        Truncates the lower C{hostbits} bits of an IP address, setting them
        to 0.
        '''
        ip = _get_ip_field(datadict, field)
        if not masklen:
            masklen = self.DEFAULT_MASKLEN[ip.version]
        else:
            masklen = int(masklen)

        datadict[field] = ip & (~((2**masklen)-1))
        self.update_mode(datadict, field, self.MODE_ANON)

    def action_ip_truncate_high(self, datadict, field, masklen=None):
        '''
        Sets the upper C{masklen} bits of an IP address to 0.
        '''
        ip = _get_ip_field(datadict, field)
        if not masklen:
            masklen = self.DEFAULT_MASKLEN[ip.version]
        else:
            masklen = int(masklen)

        datadict[field] = ip & ((2**(self.IP_BITS[ip.version]-masklen))-1)
        self.update_mode(datadict, field, self.MODE_ANON)

    def action_domain_anon(self, datadict, field, plain=0, maxparts=0):
        '''Anonymize domain names.

        Anonymous means that the domain is replaced by a random string
        without the ability to translate it back to the original, nor
        any requirements that the same domain should map to the same
        translated value.

        The algorithm will retain the last C{plain} parts of the
        domain name in its original form. It will also result in at
        most C{maxparts} parts, so that all parts beyond those
        replaced by a single element.

        '''
        _translate_domain(
            "domain_anon", datadict, field, plain, maxparts,
            lambda parts: _random_domain_part())

    def action_domain_pseudo(self, datadict, field, plain=0, maxparts=0):
        '''Pseudonymize domain names.

        Pseudonymous means that the domain is replaced by a string
        without the ability to translate it back to the original. But
        the same domain will map to the same translated value.

        The algorithm will retain the last C{plain} parts of the
        domain name in its original form. It will also result in at
        most C{maxparts} parts, so that all parts beyond those
        replaced by a single element.

        '''
        _translate_domain(
            "domain_anon", datadict, field, plain, maxparts,
            lambda parts: _pseudonymous_domain_part(parts, self.domain_seed))


def _parse_action(value):
    """Parse an action argument to a config option.

    Config options are formatted like this:

    field: action arg1 arg2 key1=arg3 key2=arg4

    This function will take the value there and return a tuple with
    three elements:

    ("action", ["arg1", "arg2"], {"key1": "arg3", "key2": "arg4"})

    """
    args = []
    kwargs = {}
    for option in value.split():
        if "=" in option:
            key, value = option.split("=", 1)
            kwargs[key] = value
        else:
            args.append(option)

    return args[0], args[1:], kwargs


def _get_ip_field(datadict, field):
    if field not in datadict:
        return None
    value = datadict[field]
    if not isinstance(value, IPAddress):
        try:
            value = IPAddress(value)
        except Exception:
            raise DataError("Field is not an IPAddress: {}"
                            .format(field))
    return value


def _split_domain(domain, maxparts):
    """Return the parts of C{domain} as a list.

    Any parts beyond C{maxparts} are returned as a single part.

    _split_domain("foo.bar.baz", 2) => ["foo.bar", "baz"]

    """
    if maxparts == 1:
        return [domain]

    parts = domain.split(".")
    prefix = parts[:-maxparts + 1]
    suffix = parts[-maxparts + 1:]
    if prefix:
        return [".".join(prefix)] + suffix
    else:
        return parts


def _translate_domain(action, datadict, field, plain, maxparts, mapper):
    if field not in datadict:
        return
    try:
        plain = int(plain)
    except ValueError:
        raise ConfigError("Bad value for argument plain in action "
                          "{}: {}".format(action, repr(plain)))
    try:
        maxparts = int(maxparts)
    except ValueError:
        raise ConfigError("Bad value for argument maxparts in action "
                          "{}: {}".format(action, repr(maxparts)))

    parts = _split_domain(datadict[field], maxparts)

    for i in range(len(parts) - plain):
        parts[i] = mapper(parts[i:])

    datadict[field] = ".".join(parts)


def _random_domain_part():
    return "".join(random.choice("abcdefghijklmnopqrstuvwxyz"  # nosec
                                 "ABCDEFGHIJKLMNOPQRSTUVWXYZ")
                   for _ in range(23))


def _pseudonymous_domain_part(parts, seed):
    if seed is None:
        raise ConfigError("Please set @domain_seed when using action "
                          "domain_pseudo")
    return hashlib.sha256((".".join(parts) + seed).encode('utf-8')).hexdigest()
