# xarfexporter.py
#
# Copyright (c) 2015 DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.message import MIMEMessage
from random import randint

from mixmod import exporter
from mixmod import utils
from mixmod.errors import DataError, ConfigError

# Structure of XARF Bulk format
#
# X-XARF BULK (multipart/mixed)
#        |
#        --- RFC822 container (message/rfc822)
#        |   of the 1st message
#        |       |
#        |       --- X-XARF PLAIN (multipart/mixed)
#        |               |
#        |               --- 1st MIME part (text/plain)
#        |               |    human readable
#        |               |
#        |               --- 2nd MIME part (text/plain)
#        |               |   YAML
#        |               |
#        |               --- 3rd MIME part (text/plain) depends on attachment
#        |                   attachment
#        |
#        --- RFC822 container (message/rfc822)
#            of the 2nd message
#                |
#                --- X-XARF PLAIN (multipart/mixed)
#                .
#                .
#                .

# Setting default values if config is None
DEFAULT_USER_AGENT = 'DFN-CERT-mixmod/0.01 (mixmod converting library 0.01)'
DEFAULT_SUBJECT = 'Flow report'
DEFAULT_REPORTER = 'mail@localhost'
XARF_VERSION = '0.2'
DEFAULT_STANDARD_TLP = 'amber'
DEFAULT_SCHEMA_URL = 'file:///private_flows_0.1.0.json'

# X-ARF field tuple (YAML field name, MixMod or internal field name)
MANDATORY_XARF_FIELDS = [
    ('Reported-From', 'xarf_reported_from'),
    ('Report-ID', 'xarf_report_id'),
    ('Category', 'xarf_category'),
    ('Report-Type', 'xarf_report_type'),
    ('User-Agent', 'xarf_user_agent'),
    ('Date', 'xarf_date'),
    ('Date-Last', 'xarf_date_last'),
    ('Source', 'xarf_src_ip'),
    ('Source-Type', 'xarf_src_type'),
    ('Destination', 'xarf_dst_ip'),
    ('Destination-Type', 'xarf_dst_type'),
    ('Source-Port', 'src_port'),
    ('Destination-Port', 'dst_port'),
    ('Protocol', 'ip_protocol_number'),
    ('Attachment', 'xarf_attachment_mimetype'),
    ('Schema-URL', 'xarf_schema_url'),
    ('Version', 'xarf_version')]

OPTIONAL_XARF_FIELDS = [
    ('Source-AS', 'src_as'),
    ('Destination-AS', 'dst_as'),
    ('Type-Of-Service', 'type_of_service'),
    ('TCP-Flags', 'tcp_flags'),
    ('SNMP-Index-In', 'snmp_index_in'),
    ('SNMP-Index-Out', 'snmp_index_out'),
    ('Byte-Count', 'byte_count'),
    ('Packet-Count', 'packet_count'),
    ('Direction', 'direction'),
    ('Occurrences', 'occurrences'),
    ('TLP', 'tlp')]


class XARFExporter(exporter.FileExporter):
    """Export mixmod data to XARF."""

    CONFIG = {
        "subject": {
            "type": "string",
            "default": DEFAULT_SUBJECT
        },
        "user_agent": {
            "type": "string",
            "default": DEFAULT_USER_AGENT,
        },
        "sender": {
            "type": "string",
            "default": None
        },
        "reporter": {
            "type": "string",
            "default": DEFAULT_REPORTER
        },
        "recipient": {
            "type": "string",
            "default": None,
        },
        "schema_url": {
            "type": "string",
            "default": DEFAULT_SCHEMA_URL
        },
        "standard_tlp": {
            "type": "string",
            "default": DEFAULT_STANDARD_TLP
        },
        "skip_dataset_on_error": {
            "type": "boolean",
            "default": "True"
        }
    }

    def __init__(self, fobj,
                 subject=DEFAULT_SUBJECT,
                 user_agent=DEFAULT_USER_AGENT,
                 sender=None,
                 reporter=DEFAULT_REPORTER,
                 recipient=None,
                 schema_url=DEFAULT_SCHEMA_URL,
                 standard_tlp=DEFAULT_STANDARD_TLP,
                 skip_dataset_on_error=True,
                 **conf):
        super(XARFExporter, self).__init__(fobj, **conf)
        self.subject = subject
        self.useragent = user_agent
        self.sender = sender
        self.reporter = reporter
        self.recipient = recipient
        self.schema_url = schema_url
        self.standard_tlp = standard_tlp
        self.skip_dataset_on_error = skip_dataset_on_error
        self.data = []

        self.message = MIMEMultipart(_subtype='mixed')
        self.message['Subject'] = self.subject
        self.message.add_header('X-XARF', 'BULK')
        self.message.add_header('Auto-Submitted', 'auto-generated')
        if self.sender is not None:
            self.message['From'] = self.sender
        if self.recipient is not None:
            self.message['To'] = self.recipient

    def put(self, data):
        try:
            self.message.attach(self._create_container(data))
        except DataError:
            if not self.skip_dataset_on_error:
                raise

    def close(self):
        # create X-XARF BULK header
        # create RFC822 container for each dataset
        self.fobj.write(self.message.as_string().encode('utf-8'))

    def _create_container(self, dataset):
        """
        Takes a MixMod dataset and returns a RFC822 container.
        """
        # if __debug__:
        #    dataset['attachment'] = 'attached data for testing purpose only.'

        # check for existence of basic fields
        basic_fields = ['ip_version', 'utc_first_packet', 'utc_last_packet',
                        'src_port', 'dst_port']
        for field in basic_fields:
            try:
                dataset[field]
            except KeyError:
                msg = 'required Mixmod field "%s" does not exist' % field
                raise DataError(msg)

        # preparations
        if dataset['ip_version'] == 4:
            try:
                dataset['xarf_src_ip'] = dataset['src_ip_v4']
                dataset['xarf_dst_ip'] = dataset['dst_ip_v4']
            except KeyError as e:
                msg = 'field does not exist in Mixmod dataset: %s' % str(e)
                raise DataError(msg)
            dataset['xarf_src_type'] = 'ipv4'
            dataset['xarf_dst_type'] = 'ipv4'
        elif dataset['ip_version'] == 6:
            try:
                dataset['xarf_src_ip'] = dataset['src_ip_v6']
                dataset['xarf_dst_ip'] = dataset['dst_ip_v6']
            except KeyError as e:
                msg = 'field does not exist in Mixmod dataset: %s' % str(e)
                raise DataError(msg)
            dataset['xarf_src_type'] = 'ipv6'
            dataset['xarf_dst_type'] = 'ipv6'
        else:
            raise DataError('IP Version {} is not supported'.format(
                dataset['ip_version']))
        if 'tlp' not in dataset:
            dataset['tlp'] = self.standard_tlp
        if 'attachment' in dataset:
            dataset['xarf_attachment_mimetype'] = 'text/plain'
        else:
            dataset['xarf_attachment_mimetype'] = 'none'
        dataset['xarf_schema_url'] = self.schema_url
        dataset['xarf_version'] = XARF_VERSION
        dataset['xarf_user_agent'] = self.useragent
        dataset['xarf_reported_from'] = self.reporter
        dataset['xarf_report_id'] = '%s-%s-%s' % (
            dataset['utc_first_packet'].strftime("%d%m%Y%H%M%S"),
            str(randint(0, 2 ** 64)),
            self.reporter)
        dataset['xarf_category'] = 'private'
        dataset['xarf_report_type'] = 'netflows'
        dataset['xarf_date'] = dataset['utc_first_packet'].strftime(
            "%a, %d %b %Y %H:%M:%S +0000")
        dataset['xarf_date_last'] = dataset['utc_last_packet'].strftime(
            "%a, %d %b %Y %H:%M:%S +0000")

        # MIME part 1 payload
        payload1 = 'Between %s and %s packets were sent ' % (
            dataset['xarf_date'],
            dataset['xarf_date_last'])
        payload1 += 'from the IP %s Port %s to the IP %s Port %s.' % (
            dataset['xarf_src_ip'],
            dataset['src_port'],
            dataset['xarf_dst_ip'],
            dataset['dst_port'])

        # MIME part 1 header
        part1 = MIMEText(payload1,
                         _subtype='plain',
                         _charset='utf-8')

        # MIME part 2 payload
        payload2 = dict()

        # setting mandatory X-ARF YAML fields
        for field in MANDATORY_XARF_FIELDS:
            self._add_yaml_payload(field, payload2, dataset, mandatory=True)

        # setting optional X-ARF YAML fields
        for field in OPTIONAL_XARF_FIELDS:
            self._add_yaml_payload(field, payload2, dataset, mandatory=False)

        # writing yaml
        yaml = '---\n'
        for key, value in payload2.items():
            yaml += "%s: %s\n" % (key, value)

        # MIME part 2 header
        part2 = MIMEText(yaml)
        del part2['Content-Type']
        part2.add_header('Content-Type', 'text/plain',
                         charset='utf-8', name='report.txt')

        # MIME part 3 payload
        if payload2['Attachment'] is not 'none':
            payload3 = dataset['attachment']
        else:
            payload3 = None

        # MIME part 3 header
        if payload2['Attachment'] is 'text/plain':
            part3 = MIMEText(payload3,
                             _subtype='plain',
                             _charset='utf-8')
        else:
            part3 = None

        # Embedded Mail Header
        embedded_mail = MIMEMultipart(_subtype='mixed')
        embedded_mail.add_header('X-XARF', 'PLAIN')
        embedded_mail.add_header('Auto-Submitted', 'auto-generated')
        embedded_mail['Subject'] = 'Abuse report about IP ' + \
                                   str(dataset['xarf_src_ip']) + \
                                   ' - ' + dataset['xarf_date']
        embedded_mail.attach(part1)
        embedded_mail.attach(part2)
        if payload2['Attachment'] is not 'none':
            embedded_mail.attach(part3)
        # RFC822 Container
        container = MIMEMessage(embedded_mail)
        container.add_header('Content-Disposition', 'attachment',
                             filename='xarf.eml')
        container.add_header('Content-Transfer-Encoding', '7bit')
        del container['Content-Type']
        container.add_header('Content-Type', 'message/rfc822', name='xarf.eml')
        return container

    def _add_yaml_payload(self, field, payload2, dataset, mandatory):
        """
        Adds a field to yaml payload.
        """
        try:
            payload2[field[0]] = dataset[field[1]]
        except KeyError:
            if mandatory:
                msg = 'mandatory field "%s" is not set' % field[1]
                raise DataError(msg)
                # if __debug__:
                #    print 'Optional field "' + field[1] + '" is not set.'
