# netflowfilter.py
#
# Copyright (c) 2011-2014, DFN-CERT Services GmbH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import datetime
import operator
import os
import re
import socket
import struct

from . import nfcap

__all__ = ['NetFlowFilter', 'T', 'ip2int', 'int2ip', 'CIDR',
           'NetFlowAggregator', 'summarize', 'timeseries']

# Default maximum length of aggregated sets
DEFAULT_MAX_LENGTH = 20


class NetFlowFilter(object):
    _filenameformat = "nfcapd.%Y%m%d%H%M"

    def __init__(self, directory, starttime=None, endtime=None,
                 _filter=None, limit=None):
        """
        Create a new NetFlowFilter object.

        directory: The directory to find nfcap files in
        starttime: The timestamp of the first netflow file to check
        endtime: The timestamp of the last netflow file to check
        limit: The maximum number of netflows to check
        """
        self._directory = directory
        self._starttime = starttime
        self._endtime = endtime
        self._limit = limit
        self._filter = _filter

    def directory(self, directory):
        return NetFlowFilter(directory, self._starttime, self._endtime,
                             self._filter, self._limit)

    def starttime(self, starttime):
        """
        Set the timestamp of the first netflow file to check.

        Returns a new object with the setting.
        """
        return NetFlowFilter(self._directory, starttime, self._endtime,
                             self._filter, self._limit)

    def endtime(self, endtime):
        """
        Set the timestamp of the last netflow file to check.

        Returns a new object with the setting.
        """
        return NetFlowFilter(self._directory, self._starttime, endtime,
                             self._filter, self._limit)

    def _maketest(self, oldtest, args, kwargs):
        """
        Combine an existing test, a test list, and keyword arguments
        into a new test.
        """
        testlist = []
        if oldtest is not None:
            testlist.append(oldtest)
        if len(args) > 0:
            testlist.extend(args)
        if len(kwargs) > 0:
            testlist.append(T(**kwargs))

        if len(testlist) > 0:
            newtest = testlist[0]
            for f in testlist[1:]:
                newtest &= f
        else:
            newtest = None

        return newtest

    def filter(self, *args, **kwargs):
        """
        Return a new object that only yields records as specified.

        Arguments are either T objects, or keywords for T object
        creation.
        """
        return NetFlowFilter(self._directory, self._starttime, self._endtime,
                             self._maketest(self._filter, args, kwargs),
                             self._limit)

    def exclude(self, *args, **kwargs):
        """
        Return a new object that yields no records as specified.

        Arguments are either T objects, or keywords for T object
        creation.
        """
        newtest = self._maketest(self._filter, args, kwargs)
        if newtest is not None:
            newtest = ~newtest
        return NetFlowFilter(self._directory, self._starttime, self._endtime,
                             newtest,
                             self._limit)

    def limit(self, limit):
        """
        Set the maximum number of netflow records to check.

        Returns a new object with the setting.
        """
        return NetFlowFilter(self._directory, self._starttime, self._endtime,
                             self._filter, limit)

    def _files(self):
        """
        Return a list of file names we will consider.
        """
        files = []
        for fname in os.listdir(self._directory):
            try:
                filetime = datetime.datetime.strptime(fname,
                                                      self._filenameformat)
            except ValueError:  # bad file format, not for us
                continue
            if self._starttime is not None and self._starttime > filetime:
                continue
            if self._endtime is not None and self._endtime < filetime:
                continue
            files.append((filetime, os.path.join(self._directory, fname)))
        files.sort()
        return files

    def maxflows(self):
        """
        Return the number of flows the object would check.

        This only depends on the start and end time.
        """
        count = 0
        for filetime, fullname in self._files():
            nf = nfcap.NFCap(file(fullname))
            count += nf.numflows["total"]
        return count

    def __iter__(self):
        """
        Iterate over all netflows, yielding only those requested.
        """
        counter = 0
        for filetime, fullname in self._files():
            nf = nfcap.NFCap(file(fullname))
            for record in nf.iter_records():
                if self._limit:
                    counter += 1
                    if counter > self._limit:
                        return
                if self._filter is None or self._filter(record):
                    yield (filetime, record)

    def __getitem__(self, item):
        if isinstance(item, slice):
            return self._getslice(item)
        else:
            i = 0
            for elt in self:
                if i == item:
                    return elt
                i += 1
            raise IndexError("Not enough netflows")

    def _getslice(self, item):
        i = 0
        if (
                item.start is not None and item.start < 0 or
                item.stop is not None and item.stop < 0
        ):
            raise RuntimeError("NetFlowFilter does not support "
                               "negative indices")
        next = item.start if item.start is not None else 0
        for elt in self:
            if item.stop is not None and i >= item.stop:
                return
            elif i == next:
                yield elt
                if item.step is None:
                    next += 1
                else:
                    next += item.step
            i += 1


##################################################################
# Efficient T objects

# These compile to Python bytecode for speed

class T(object):
    def __init__(self, tree=None, **kwargs):
        """
        Create a new test object.

        Arguments are keywords of the form attribute__operation, where
        attribute can be any attribute of a netflow record, and
        operation can be any of the operations in the OPERATIONS
        dictionary. The operation can be left out and defaults to
        'eq'.

        All tests must be satisfied.

        ops are existing operations. All operations in ops must yield
        true in addition to the tests in kwargs to make this operation
        yield true.

        The current implementation only allows objects as argument
        values whose repr() compiles back to the original object.
        """
        if tree is None:
            if len(kwargs) == 1:
                self.tree = OpTestNode(*self._makeop(list(kwargs.items())[0]))
            else:
                self.tree = AndTestNode(*[OpTestNode(*self._makeop(arg))
                                          for arg in kwargs.items()])
        else:
            assert len(kwargs) == 0
            self.tree = tree
        self.opfun = self.compile()

    def __repr__(self):
        return "<T %s>" % (self.tree.python(),)

    def __or__(self, othertest):
        """
        Return a new test object that tests successfully if this test
        or othertest is successful.
        """
        return T(OrTestNode(self.tree, othertest.tree))

    def __and__(self, othertest):
        """
        Return a new test object that tests successfully only if this
        test and othertest are successful.
        """
        return T(AndTestNode(self.tree, othertest.tree))

    def __invert__(self):
        return T(NotTestNode(self.tree))

    def __call__(self, obj):
        """
        The actual test of obj. Returns True or False.
        """
        return self.opfun(obj)

    def compile(self):
        expr = "lambda rec: %s" % (self.tree.python(),)
        return eval(compile(expr, "<string>", "eval"))

    def _makeop(self, optuple):
        """
        Create a new test op for this keyword argument.

        Splits the keyword at __ and returns a function that does this
        test.
        """
        key, value = optuple
        if "__" in key:
            (field, op) = key.split("__", 1)
        else:
            (field, op) = (key, 'eq')
        return (op, field, value)


class TestNode(object):
    MAX_FIELD_ORDER = 100

    def __init__(self):
        pass

    def python(self):
        return "False"

    def order(self):
        return self.MAX_FIELD_ORDER


class OpTestNode(TestNode):
    IDENTIFIER_ESCAPE_RX = re.compile("[^A-Za-z0-9_]")

    OPERATIONS = {'gt': lambda field, arg: "rec.%s > %r" % (field, arg),
                  'lt': lambda field, arg: "rec.%s < %r" % (field, arg),
                  'gte': lambda field, arg: "rec.%s >= %r" % (field, arg),
                  'lte': lambda field, arg: "rec.%s <= %r" % (field, arg),
                  'eq': lambda field, arg: "rec.%s == %r" % (field, arg),
                  'ne': lambda field, arg: "rec.%s != %r" % (field, arg),
                  'in': lambda field, arg: "rec.%s in %r" % (field, arg),
                  'in_cidr': lambda field, arg: ("(%r == rec.%s & %r and rec.ipversion == %r)" %
                                                 (arg.ip, field, arg.mask, arg.ipversion)),
                  'between': lambda field, arg: ("rec.%s >= %r and "
                                                 "rec.%s <= %r" %
                                                 (field, arg[0],
                                                  field, arg[1])),
                  'bitmask': lambda field, arg: ("(rec.%s & %r) == %r" %
                                                 (field, arg, arg))
                  }

    # Ordering for the fields. Smaller means check earlier.
    FIELDORDER = {'input': 1,
                  'output': 1,
                  'srcport': 2,
                  'dstport': 2,
                  'prot': 3,
                  'tos': 3,
                  'icmpcode': 3,
                  'icmptype': 3,
                  'srcip': 4,
                  'dstip': 4,
                  'tcp_flags': 5,
                  'srccountry': 6,
                  'dstcountry': 6,
                  'start': 7,
                  'end': 8}

    def __init__(self, op, field, value):
        self.op = op
        self.field = self.IDENTIFIER_ESCAPE_RX.sub("", field)
        self.value = value

    def order(self):
        order = self.FIELDORDER.get(self.field, None)
        if order is None:
            return super(OpTestNode, self).order()
        else:
            return order

    def python(self):
        if self.op == 'in' and isinstance(self.value, CIDR):
            comp = 'in_cidr'
        else:
            comp = self.op
        return "%s" % self.OPERATIONS[comp](self.field, self.value)


class AndTestNode(TestNode):
    def __init__(self, *args):
        self.args = []
        for arg in args:
            if isinstance(arg, AndTestNode):
                self.args.extend(arg.args)
            else:
                self.args.append(arg)

    def python(self):
        if len(self.args) == 0:
            return "True"
        return " and ".join(["(%s)" % (arg.python(),)
                             for arg in sorted(self.args,
                                               key=lambda node: node.order()
                                               )])


class OrTestNode(TestNode):
    def __init__(self, *args):
        self.args = []
        for arg in args:
            if isinstance(arg, OrTestNode):
                self.args.extend(arg.args)
            else:
                self.args.append(arg)

    def python(self):
        if len(self.args) == 0:
            return "False"
        return " or ".join(["(%s)" % (arg.python(),)
                            for arg in sorted(self.args,
                                              key=lambda node: node.order()
                                              )])


class NotTestNode(TestNode):
    def __init__(self, arg):
        self.arg = arg

    def python(self):
        return "not (%s)" % (self.arg.python(),)


##################################################################
# Helper functions

def ip2int(ipstring):
    """
    Parse an IP string into the corresponding integer.

    Works for both IPv4 and IPv6.
    """
    if ":" in ipstring:
        bytes = socket.inet_pton(socket.AF_INET6, ipstring)
        (high, low) = struct.unpack("!QQ", bytes)
        return high * (2 ** 64) + low
    else:
        bytes = socket.inet_pton(socket.AF_INET, ipstring)
        return struct.unpack("!I", bytes)[0]


def int2ip(number):
    """
    Return the string representation of an IP integer.

    Works for both IPv4 and IPv6
    """
    # Basically, we assume that an IP in the range of IPv4 is actually
    # an IPv4 address. Anything else is IPv6. This is not perfect, but
    # it should work pretty much all the time as the 0::0 net is
    # reserved.
    if number > 2 ** 32:
        # IPv6
        high = number >> 64
        low = number & ((2 ** 64) - 1)
        bytes = struct.pack("!QQ", high, low)
        return socket.inet_ntop(socket.AF_INET6, bytes)
    else:
        # IPv4
        bytes = struct.pack("!I", number)
        return socket.inet_ntop(socket.AF_INET, bytes)


class CIDR(object):
    def __init__(self, cidr):
        """
        Create a new CIDR object.

        The CIDR string may leave out the /32 or /128.
        """
        self.name = cidr
        if ":" in cidr:
            self.size = 128
            self.ipversion = 6
        else:
            self.size = 32
            self.ipversion = 4
        if "/" in cidr:
            (ipstring, bits) = cidr.split("/", 1)
            self.ip = ip2int(ipstring)
            self.bits = int(bits)
        else:
            self.ip = ip2int(cidr)
            self.bits = self.size
        self.mask = (2 ** self.size - 1) - (2 ** (self.size - self.bits) - 1)
        self.ip = self.ip & self.mask

    def __eq__(self, other):
        return (self.ip == other.ip and
                self.mask == other.mask)

    def __contains__(self, ip):
        return self.ip == ip & self.mask

    def __repr__(self):
        return "<CIDR %s>" % self.name


class NetFlowAggregator(object):
    def __init__(self, sequence, max_length=DEFAULT_MAX_LENGTH,
                 _aggregate_on=[]):
        """
        Create a new netflow aggregator object.

        sequence: Any sequence of netflow records.
        _aggregate_on: List of fields to aggregate on
        """
        self.sequence = sequence
        self._max_length = max_length
        self._aggregate_on = _aggregate_on

    def __iter__(self):
        results = {}
        for filetime, record in self.sequence:
            key = self.get_key(record)
            value = results.get(key, None)
            if value is None:
                value = FakeRecord(self._max_length,
                                   _overrides=[fieldname for (fieldname,
                                                              ignored)
                                               in self._aggregate_on
                                               if fieldname in ['numpackets',
                                                                'numbytes']])
                results[key] = value
            value.add(record)
        for rec in sorted(results.values(), key=lambda rec: rec.first):
            yield rec

    def get_key(self, record):
        """
        Return a key that is the same for any record that's supposed
        to be aggregated.
        """
        result = []
        for key, value in self._aggregate_on:
            if value is True:
                result.append(getattr(record, key))
            elif isinstance(value, CIDR):
                result.append(getattr(record, key) & value.mask)
            else:
                raise RuntimeError("Bad key value: %s" % (value,))
        return tuple(result)

    def aggregate_on(self, **kwargs):
        return NetFlowAggregator(self.sequence,
                                 self._max_length,
                                 self._aggregate_on + list(kwargs.items()))


class FakeRecord(object):
    def __init__(self, max_length=DEFAULT_MAX_LENGTH, _overrides=[]):
        self._max_length = max_length
        self.numflows = 0
        self._fields = []
        for field, op in [('first', self.add_min),
                          ('last', self.add_max),
                          ('tcp_flags', self.add_or),
                          ('prot', self.add_set),
                          ('srcport', self.add_set),
                          ('dstport', self.add_set),
                          ('numpackets', self.add_sum),
                          ('numbytes', self.add_sum),
                          ('srcip', self.add_set),
                          ('dstip', self.add_set),
                          ('srccountry', self.add_set),
                          ('dstcountry', self.add_set),
                          ('start', self.add_set),
                          ('end', self.add_set),
                          ('input', self.add_set),
                          ('output', self.add_set),
                          ('srcas', self.add_set),
                          ('dstas', self.add_set),
                          ('ipversion', self.add_set),
                          ('icmptypecode', self.add_set)]:
            if field in _overrides:
                self._fields.append((field, self.add_equal))
            else:
                self._fields.append((field, op))

    def __getattr__(self, fieldname):
        """
        Unknown fields have None as a value.
        """
        return None

    def add(self, record):
        for fieldname, op in self._fields:
            op(fieldname, record)
        self.numflows += 1

    def add_op(self, fieldname, record, op):
        value = getattr(record, fieldname)
        if value is not None:
            cur_value = getattr(self, fieldname)
            if cur_value is None:
                setattr(self, fieldname, value)
            else:
                setattr(self, fieldname, op(cur_value, value))

    def add_equal(self, fieldname, record):
        self.add_op(fieldname, record, lambda a, b: a)

    def add_or(self, fieldname, record):
        self.add_op(fieldname, record, operator.__or__)

    def add_sum(self, fieldname, record):
        self.add_op(fieldname, record, operator.__add__)

    def add_max(self, fieldname, record):
        self.add_op(fieldname, record, max)

    def add_min(self, fieldname, record):
        self.add_op(fieldname, record, min)

    def add_set(self, fieldname, record):
        if fieldname == 'icmptypecode' and hasattr(record, 'icmptype'):
            value = (record.icmptype, record.icmpcode)
        else:
            value = getattr(record, fieldname)
        if value is not None:
            cur_value = getattr(self, fieldname)
            if cur_value is None:
                cur_value = set()
                setattr(self, fieldname, cur_value)
            if len(cur_value) < self._max_length:
                cur_value.add(value)


def summarize(nff, fieldname):
    """
    Summarize all fields named FIELDNAME in the net flow filter.

    Returns a list of (interface, count) tuples
    """
    result = {}
    total = 0
    for filetime, record in nff:
        value = getattr(record, fieldname)
        if value is not None:
            total += value
            for iface in [record.input, record.output]:
                if iface in result:
                    result[iface] += value
                else:
                    result[iface] = value
    return total, list(result.items())


def timeseries(nff, fieldnames):
    """
    Summarize all fields in FIELDNAMES in the net flow filter by minute.

    Returns a list of (timestamp, mapping) tuples, where mapping maps
    the field name to the sum.
    """
    result = []
    lastts = None
    current = None
    for filetime, record in nff:
        if filetime != lastts:
            current = dict((name, 0) for name in fieldnames)
            result.append((filetime, current))
            lastts = filetime
        for field in fieldnames:
            current[field] += getattr(record, field)
    return result
